"""Contains the Mesh class used by Archive."""

from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union, cast
import warnings

import numpy as np
import numpy.typing as npt
from pyvista import PolyData, UnstructuredGrid

from mapdl_archive import _reader, _relaxmidside
from mapdl_archive.elements import ETYPE_MAP

COMP_DICT = Dict[str, npt.NDArray[np.int32]]

INVALID_ALLOWABLE_TYPES = TypeError(
    "`allowable_types` must be an array of ANSYS element types from 1 and 300"
)

# map MESH200 elements to a mapdl_archive/VTK element type (see elements.py)
MESH200_MAP: Dict[int, int] = {
    0: 2,  # line
    1: 2,  # line
    2: 2,  # line
    3: 2,  # line
    4: 3,  # triangle
    5: 3,  # triangle
    6: 3,  # quadrilateral
    7: 3,  # quadrilateral
    8: 5,  # tetrahedron with 4 nodes
    9: 5,  # tetrahedron with 10 nodes
    10: 4,  # hex with 8 nodes
    11: 4,
}  # hex with 8 nodes

SHAPE_MAP: Dict[int, str] = {  # from ELIST definition
    0: "",
    1: "LINE",
    2: "PARA",
    3: "ARC ",
    4: "CARC",
    5: "",
    6: "TRIA",
    7: "QUAD",
    8: "TRI6",
    9: "QUA8",
    10: "POIN",
    11: "CIRC",
    12: "",
    13: "",
    14: "CYLI",
    15: "CONE",
    16: "SPHE",
    17: "",
    18: "",
    19: "PILO",
}
# element type to VTK conversion function call map
# 0: skip
# 1: Point
# 2: Line (linear or quadratic)
# 3: Shell
# 4: 3D Solid (Hexahedral, wedge, pyramid, tetrahedral)
# 5: Tetrahedral
# 6: Line (always linear)
TARGE170_MAP = {
    "TRI": 3,  # 3-Node Triangle
    "QUAD": 3,  # 4-Node Quadrilateral
    "CYLI": 0,  # Not supported (NS)  # Cylinder
    "CONE": 0,  # NS  # Cone
    "TRI6": 3,  # 6-Node triangle
    "SPHE": 0,  # NS  # Sphere
    "PILO": 1,  # Pilot Node
    "QUAD8": 3,  # 8-Node Quadrilateral
    "LINE": 2,  # Line
    "PARA": 2,  # Parabola
    "POINT": 1,  # Point
}


def unique_rows(a: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Return unique rows of an array and the indices of those rows."""
    if not a.flags.c_contiguous:
        a = np.ascontiguousarray(a)

    b = a.view(np.dtype((np.void, a.dtype.itemsize * a.shape[1])))
    _, idx, idx2 = np.unique(b, True, True)

    return a[idx], idx, idx2


class Mesh:
    """Common class between Archive, and result mesh."""

    def __init__(
        self,
        nnum: npt.NDArray[np.int32],
        nodes: npt.NDArray[np.float64],
        elem: npt.NDArray[np.int32],
        elem_off: npt.NDArray[np.int32],
        ekey: npt.NDArray[np.int32],
        rnum: npt.NDArray[np.int64],
        node_comps: COMP_DICT = {},
        elem_comps: COMP_DICT = {},
        rdat: List[List[float]] = [],
        keyopt: Dict[int, List[List[int]]] = {},
    ):
        """Initialize the mesh."""
        self._etype: Optional[npt.NDArray[np.int32]] = (
            None  # internal element type reference
        )
        self._grid: Optional[UnstructuredGrid] = None
        self._surf_cache: Optional[PolyData] = None  # cached external surface
        self._enum: Optional[npt.NDArray[np.int32]] = None  # cached element numbering
        self._etype_cache: Optional[npt.NDArray[np.int32]] = (
            None  # cached ansys ETYPE num
        )
        self._rcon: Optional[npt.NDArray[np.int32]] = (
            None  # ansys element real constant
        )
        self._mtype: Optional[npt.NDArray[np.int32]] = (
            None  # cached ansys material type
        )
        self._node_angles: Optional[npt.NDArray[np.float64]] = (
            None  # cached node angles
        )
        self._node_coord: Optional[npt.NDArray[np.float64]] = (
            None  # cached node coordinates
        )
        self._cached_elements: Optional[List[npt.NDArray[np.int32]]] = None
        self._secnum: Optional[npt.NDArray[np.int32]] = None  # cached section number
        self._esys: Optional[npt.NDArray[np.int32]] = (
            None  # cached element coordinate system
        )
        self._etype_id: Optional[npt.NDArray[np.int32]] = None  # cached element type id

        # Always set on init
        self._nnum: npt.NDArray[np.int32] = nnum
        self._nodes: npt.NDArray[np.float64] = nodes
        self._elem: npt.NDArray[np.int32] = elem
        self._elem_off: npt.NDArray[np.int32] = elem_off
        self._ekey: npt.NDArray[np.int32] = ekey

        # optional
        self._node_comps: COMP_DICT = node_comps
        self._elem_comps: COMP_DICT = elem_comps
        self._rdat: List[List[float]] = rdat
        self._rnum: npt.NDArray[np.int64] = rnum
        self._keyopt: Dict[int, List[List[int]]] = keyopt
        self._tshape: Optional[npt.NDArray[np.int32]] = None
        self._tshape_key: Optional[npt.NDArray[np.int32]] = None

    @property
    def _surf(self) -> PolyData:
        """Return the external surface."""
        if self._surf_cache is None:
            if self._grid is None:
                raise AttributeError("Missing grid.")
            self._surf_cache = self._grid.extract_surface()
        return self._surf_cache

    @property
    def _has_nodes(self) -> bool:
        """Return ``True`` when has nodes."""
        return bool(len(self.nodes))

    @property
    def _has_elements(self) -> bool:
        """Return ``True`` when geometry has elements."""
        if self._elem is None:
            return False

        if isinstance(self._elem, np.ndarray):
            return bool(self._elem.size)

        return bool(len(self._elem))

    def _parse_vtk(
        self,
        allowable_types: Optional[List[Union[int, str]]] = None,
        force_linear: bool = False,
        null_unallowed: bool = False,
        fix_midside: bool = True,
        additional_checking: bool = False,
    ) -> UnstructuredGrid:
        """Convert raw ANSYS nodes and elements to an UnstructuredGrid.

        Parameters
        ----------
        fix_midside : bool, optional
            Adds additional midside nodes when ``True``. When ``False``,
            missing ANSYS cells will simply point to the first node.

        """
        if not self._has_nodes or not self._has_elements:
            # warnings.warn('Missing nodes or elements.  Unable to parse to vtk')
            return

        etype_map = ETYPE_MAP
        if allowable_types is not None:
            try:
                allowable_types_arr: npt.NDArray[np.int_] = np.asarray(allowable_types)
            except Exception as e:
                warnings.warn(f"{e}")
                raise INVALID_ALLOWABLE_TYPES

            if not issubclass(allowable_types_arr.dtype.type, np.integer):
                raise TypeError("Element types must be an integer array-like")

            if allowable_types_arr.min() < 1 or allowable_types_arr.max() > 300:
                raise INVALID_ALLOWABLE_TYPES

            etype_map = np.zeros_like(ETYPE_MAP)
            etype_map[allowable_types] = ETYPE_MAP[allowable_types_arr]

        # ANSYS element type to VTK map
        type_ref = np.empty(2 << 16, np.int32)  # 131072
        type_ref[self._ekey[:, 0]] = etype_map[self._ekey[:, 1]]

        if allowable_types is None or 200 in allowable_types:
            for etype_ind, etype in self._ekey:
                # MESH200
                if etype == 200 and etype_ind in self.key_option:
                    # keyoption 1 contains various cell types
                    # map them to the corresponding type (see elements.py)
                    mapped = MESH200_MAP[self.key_option[etype_ind][0][1]]
                    type_ref[etype_ind] = mapped

                # TARGE170 specifics
                if etype == 170:
                    # edge case where missing element within the tshape_key
                    if etype_ind not in self.tshape_key:  # pragma: no cover
                        continue
                    tshape_num = cast(int, self.tshape_key[etype_ind])
                    if (
                        tshape_num >= 19
                    ):  # weird bug when 'PILO' can be 99 instead of 19.
                        tshape_num = 19
                    tshape_label = SHAPE_MAP[tshape_num]
                    type_ref[etype_ind] = TARGE170_MAP.get(tshape_label, 0)

        offset, celltypes, cells = _reader.ans_vtk_convert(
            self._elem, self._elem_off, type_ref, self.nnum, True
        )  # for reset_midside

        nodes, angles, nnum = self.nodes, self.node_angles, self.nnum

        # fix missing midside
        if np.any(cells == -1):
            if fix_midside:
                nodes, angles, nnum = fix_missing_midside(
                    cells, nodes, celltypes, offset, angles, nnum
                )
            else:
                cells[cells == -1] = 0

        if additional_checking:
            cells[cells < 0] = 0
            # cells[cells >= nodes.shape[0]] = 0  # fails when n_nodes < 20

        grid = UnstructuredGrid(cells, celltypes, nodes, deep=True)

        # Store original ANSYS element and node information
        grid.point_data["ansys_node_num"] = nnum
        grid.cell_data["ansys_elem_num"] = self.enum
        grid.cell_data["ansys_real_constant"] = self.elem_real_constant
        grid.cell_data["ansys_material_type"] = self.material_type
        grid.cell_data["ansys_etype"] = self._ans_etype
        grid.cell_data["ansys_elem_type_num"] = self.etype

        # add components
        # Add element components to unstructured grid
        for key, item in self.element_components.items():
            mask = np.in1d(self.enum, item, assume_unique=True)
            grid.cell_data[key] = mask

        # Add node components to unstructured grid
        for key, item in self.node_components.items():
            mask = np.in1d(nnum, item, assume_unique=True)
            grid.point_data[key] = mask

        # store node angles
        if angles is not None:
            if angles.shape[1] == 3:
                grid.point_data["angles"] = angles

        if not null_unallowed:
            grid = grid.extract_cells(grid.celltypes != 0)

        if force_linear:
            # only run if the grid has points or cells
            if grid.n_points:
                grid = grid.linear_copy()

        # map over element types
        # Add tracker for original node numbering
        ind = np.arange(grid.n_points)
        grid.point_data["origid"] = ind
        grid.point_data["VTKorigID"] = ind
        return grid

    @property
    def key_option(self) -> Dict[int, List[List[int]]]:
        """Return additional key options for element types.

        Examples
        --------
        >>> import mapdl_archive
        >>> from mapdl_archive import examples
        >>> archive = mapdl_archive.Archive(examples.hexarchivefile)
        >>> archive.key_option
        {1: [[1, 11]]}
        """
        return self._keyopt

    @property
    def material_type(self) -> npt.NDArray[np.int32]:
        """Return the material type index of each element in the archive.

        Examples
        --------
        >>> import mapdl_archive
        >>> from mapdl_archive import examples
        >>> archive = mapdl_archive.Archive(examples.hexarchivefile)
        >>> archive.material_type
        array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
               1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
               1, 1, 1, 1], dtype=int32)
        """
        # FIELD 0 : material reference number
        if self._mtype is None:
            self._mtype = self._elem[self._elem_off[:-1]]
        return self._mtype

    @property
    def element_components(self) -> COMP_DICT:
        """Return Element components for the archive.

        Output is a dictionary of element components. Each entry is an array of
        MAPDL element numbers corresponding to the element component. The keys
        are element component names.

        Examples
        --------
        >>> import mapdl_archive
        >>> from mapdl_archive import examples
        >>> archive = mapdl_archive.Archive(examples.hexarchivefile)
        >>> archive.element_components
        {'ECOMP1 ': array([17, 18, 21, 22, 23, 24, 25, 26, 27, 28, 29,
                           30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40],
                           dtype=int32),
        'ECOMP2 ': array([ 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13,
                          14, 15, 16, 17, 18, 19, 20, 23, 24], dtype=int32)}
        """
        return self._elem_comps

    @property
    def node_components(self) -> COMP_DICT:
        """Return the node components for the archive.

        Output is a dictionary of node components. Each entry is an array of
        MAPDL node numbers corresponding to the node component. The keys are
        node component names.

        Examples
        --------
        >>> import mapdl_archive
        >>> from mapdl_archive import examples
        >>> archive = mapdl_archive.Archive(examples.hexarchivefile)
        >>> archive.node_components
        {'NCOMP2  ': array([  1,   2,   3,   4,   5,   6,   7,   8,
                             14, 15, 16, 17, 18, 19, 20, 21, 43, 44,
                             62, 63, 64, 81, 82, 90, 91, 92, 93, 94,
                             118, 119, 120, 121, 122, 123, 124, 125,
                             126, 137, 147, 148, 149, 150, 151, 152,
                             153, 165, 166, 167, 193, 194, 195, 202,
                             203, 204, 205, 206, 207, 221, 240, 258,
                             267, 268, 276, 277, 278, 285, 286, 287,
                             304, 305, 306, 313, 314, 315, 316
                             ], dtype=int32),
        ...,
        }
        """
        return self._node_comps

    @property
    def elem_real_constant(self) -> npt.NDArray[np.int32]:
        """Return the real constant reference for each element.

        Use the data within ``rlblock`` and ``rlblock_num`` to get the
        real constant datat for each element.

        Examples
        --------
        >>> import mapdl_archive
        >>> from mapdl_archive import examples
        >>> archive = mapdl_archive.Archive(examples.hexarchivefile)
        >>> archive.elem_real_constant
        array([ 1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,
                1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,
                1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,
                ...,
                1,  1,  1,  1,  1,  1,  1,  1,  1,  1, 61, 61, 61, 61,
               61, 61, 61, 61, 61, 61, 61, 61, 61, 61, 61, 61, 61, 61,
               61], dtype=int32)

        """
        # FIELD 2 : real constant reference number
        if self._rcon is None:
            self._rcon = self._elem[self._elem_off[:-1] + 2]
        return self._rcon

    @property
    def etype(self) -> npt.NDArray[np.int32]:
        """Return the element type of each element.

        Examples
        --------
        >>> import mapdl_archive
        >>> from mapdl_archive import examples
        >>> archive = mapdl_archive.Archive(examples.hexarchivefile)
        >>> archive.etype
        array([ 45,  45,  45,  45,  45,  45,  45,  45,  45,  45,  45,
                45,  45,  45,  45,  45,  45,  45,  45,  92,  92,  92,
                92,  92,  92,  92,  92,  92,  92,  92,  92,  92,  92,
                ...,
                92,  92,  92,  92,  92, 154, 154, 154, 154, 154, 154,
               154, 154, 154, 154, 154, 154, 154, 154, 154, 154, 154,
               154], dtype=int32)

        Notes
        -----
        Element types are listed below. Please see the APDL Element Reference
        for more details:

        https://www.mm.bme.hu/~gyebro/files/vem/ansys_14_element_reference.pdf
        """
        if self._etype is None:
            arr = np.empty(self._ekey[:, 0].max() + 1, np.int32)
            arr[self._ekey[:, 0]] = self._ekey[:, 1]
            self._etype = arr[self._ans_etype]
        return self._etype

    @property
    def _ans_etype(self) -> npt.NDArray[np.int32]:
        """Return field 1, the element type number."""
        if self._etype_cache is None:
            self._etype_cache = self._elem[self._elem_off[:-1] + 1]
        return self._etype_cache

    @property
    def section(self) -> npt.NDArray[np.int32]:
        """Return the section number.

        Examples
        --------
        >>> import mapdl_archive
        >>> from mapdl_archive import examples
        >>> archive = mapdl_archive.Archive(examples.hexarchivefile)
        >>> archive.section
        array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
               1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
               1, 1], dtype=int32)
        """
        if self._secnum is None:
            self._secnum = self._elem[self._elem_off[:-1] + 3]  # FIELD 3
        return self._secnum

    def element_coord_system(self) -> npt.NDArray[np.int32]:
        """Return the element coordinate system number.

        Examples
        --------
        >>> import mapdl_archive
        >>> from mapdl_archive import examples
        >>> archive = mapdl_archive.Archive(examples.hexarchivefile)
        >>> archive.element_coord_system
        array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
               0, 0], dtype=int32)
        """
        if self._esys is None:
            self._esys = self._elem[self._elem_off[:-1] + 4]  # FIELD 4
        return self._esys

    @property
    def elem(self) -> List[npt.NDArray[np.int32]]:
        """Return the list of elements containing raw element information.

        Each element contains 10 items plus the nodes belonging to the
        element.  The first 10 items are:

        - FIELD 0 : material reference number
        - FIELD 1 : element type number
        - FIELD 2 : real constant reference number
        - FIELD 3 : section number
        - FIELD 4 : element coordinate system
        - FIELD 5 : death flag (0 - alive, 1 - dead)
        - FIELD 6 : solid model reference
        - FIELD 7 : coded shape key
        - FIELD 8 : element number
        - FIELD 9 : base element number (applicable to reinforcing elements only)
        - FIELDS 10 - 30 : The nodes belonging to the element in ANSYS numbering.

        Examples
        --------
        >>> import mapdl_archive
        >>> from mapdl_archive import examples
        >>> archive = mapdl_archive.Archive(examples.hexarchivefile)
        >>> archive.elem
        [array([  1,   4,  19,  15,  63,  91, 286, 240,   3,  18,  17,
                 16,  81, 276, 267, 258,  62,  90, 285, 239],
         array([  4,   2,   8,  19,  91,  44, 147, 286,   5,   7,  21,
                 18, 109, 137, 313, 276,  90,  43, 146, 285],
         array([ 15,  19,  12,  10, 240, 286, 203, 175,  17,  20,  13,
                 14, 267, 304, 221, 230, 239, 285, 202, 174],
        ...
        """
        if self._cached_elements is None:
            self._cached_elements = np.split(self._elem, self._elem_off[1:-1])
        return self._cached_elements

    @property
    def enum(self) -> npt.NDArray[np.int32]:
        """Return the MAPDl element numbers.

        Examples
        --------
        >>> import mapdl_archive
        >>> from mapdl_archive import examples
        >>> archive = mapdl_archive.Archive(examples.hexarchivefile)
        >>> archive.enum
        array([    1,     2,     3, ...,  9998,  9999, 10000])
        """
        if self._enum is None:
            self._enum = self._elem[self._elem_off[:-1] + 8]
        return self._enum

    @property
    def nnum(self) -> npt.NDArray[np.int32]:
        """Return the array of node numbers.

        Examples
        --------
        >>> import mapdl_archive
        >>> from mapdl_archive import examples
        >>> archive = mapdl_archive.Archive(examples.hexarchivefile)
        >>> archive.nnum
        array([    1,     2,     3, ..., 19998, 19999, 20000])
        """
        return self._nnum

    @property
    def ekey(self) -> npt.NDArray[np.int32]:
        """Return the element type key.

        Array containing element type numbers in the first column and
        the element types (like SURF154) in the second column.

        Examples
        --------
        >>> import mapdl_archive
        >>> from mapdl_archive import examples
        >>> archive = mapdl_archive.Archive(examples.hexarchivefile)
        >>> archive.ekey
        array([[  1,  45],
               [  2,  95],
               [  3,  92],
               [ 60, 154]], dtype=int32)
        """
        return self._ekey

    @property
    def rlblock(self) -> List[List[float]]:
        """Return the real constant data from the RLBLOCK.

        Examples
        --------
        >>> import mapdl_archive
        >>> from mapdl_archive import examples
        >>> archive = mapdl_archive.Archive(examples.hexarchivefile)
        >>> archive.rlblock
        [[0.   , 0.   , 0.   , 0.   , 0.   , 0.   , 0.02 ],
         [0.   , 0.   , 0.   , 0.   , 0.   , 0.   , 0.01 ],
         [0.   , 0.   , 0.   , 0.   , 0.   , 0.   , 0.005],
         [0.   , 0.   , 0.   , 0.   , 0.   , 0.   , 0.005]]
        """
        return self._rdat

    @property
    def rlblock_num(self) -> npt.NDArray[np.int64]:
        """Return the indices from the real constant data.

        Examples
        --------
        >>> import mapdl_archive
        >>> from mapdl_archive import examples
        >>> archive = mapdl_archive.Archive(examples.hexarchivefile)
        >>> archive.rnum
        array([60, 61, 62, 63])
        """
        return self._rnum

    @property
    def nodes(self) -> npt.NDArray[np.float64]:
        """Return the array of nodes of the mesh.

        Examples
        --------
        >>> import mapdl_archive
        >>> from mapdl_archive import examples
        >>> archive = mapdl_archive.Archive(examples.hexarchivefile)
        >>> archive.nodes
        [[0.   0.   0.  ]
         [1.   0.   0.  ]
         [0.25 0.   0.  ]
         ...,
         [0.75 0.5  3.5 ]
         [0.75 0.5  4.  ]
         [0.75 0.5  4.5 ]]
        """
        if self._node_coord is None:
            self._node_coord = np.ascontiguousarray(self._nodes[:, :3])
        return self._node_coord

    @property
    def node_angles(self) -> npt.NDArray[np.float64]:
        """Return the node angles from the archive file.

        Examples
        --------
        >>> import mapdl_archive
        >>> from mapdl_archive import examples
        >>> archive = mapdl_archive.Archive(examples.hexarchivefile)
        >>> archive.nodes
        [[0.   0.   0.  ]
         [0.   0.   0.  ]
         [0.   0.   0.  ]
         ...,
         [0.   0.   0.  ]
         [0.   0.   0.  ]
         [0.   0.   0.  ]]

        """
        if self._node_angles is None:
            self._node_angles = np.ascontiguousarray(self._nodes[:, 3:])
        return self._node_angles

    def __repr__(self) -> str:
        """Return the representation of the mesh."""
        txt: str = "ANSYS Mesh\n"
        txt += f"  Number of Nodes:              {len(self.nnum)}\n"
        txt += f"  Number of Elements:           {len(self.enum)}\n"
        txt += f"  Number of Element Types:      {len(self.ekey)}\n"
        txt += f"  Number of Node Components:    {len(self.node_components)}\n"
        txt += f"  Number of Element Components: {len(self.element_components)}\n"
        return txt

    def save(
        self,
        filename: Union[str, Path],
        binary: bool = True,
        force_linear: bool = False,
        allowable_types: Optional[List[Union[int, str]]] = None,
        null_unallowed: bool = False,
    ) -> None:
        """Save the geometry as a vtk file.

        Parameters
        ----------
        filename : str, pathlib.Path
            Filename of output file. Writer type is inferred from
            the extension of the filename.
        binary : bool, optional
            If ``True``, write as binary, else ASCII.
        force_linear : bool, optional
            This parser creates quadratic elements if available.  Set
            this to True to always create linear elements.  Defaults
            to False.
        allowable_types : list, optional
            Allowable element types.  Defaults to all valid element
            types in ``mapdl_archive.elements.valid_types``

            See ``help(mapdl_archive.elements)`` for available element types.
        null_unallowed : bool, optional
            Elements types not matching element types will be stored
            as empty (null) elements.  Useful for debug or tracking
            element numbers.  Default False.

        Examples
        --------
        >>> geom.save("mesh.vtk")

        Notes
        -----
        Binary files write much faster than ASCII and have a smaller
        file size.

        """
        grid: UnstructuredGrid = self._parse_vtk(
            allowable_types=allowable_types,
            force_linear=force_linear,
            null_unallowed=null_unallowed,
        )
        grid.save(str(filename), binary=binary)
        return None

    @property
    def n_node(self) -> int:
        """Return the number of nodes."""
        if not self._has_nodes:
            return 0
        return self.nodes.shape[0]

    @property
    def n_elem(self) -> int:
        """Return the number of nodes."""
        if not self._has_elements:
            return 0
        return len(self.enum)

    @property
    def et_id(self) -> npt.NDArray[np.int32]:
        """Element type id (ET) for each element."""
        if self._etype_id is None:
            etype_elem_id = self._elem_off[:-1] + 1
            self._etype_id = self._elem[etype_elem_id]
        return self._etype_id

    @property
    def tshape(self) -> npt.NDArray[np.int32]:
        """Tshape of contact elements."""
        if self._tshape is None:
            shape_elem_id = self._elem_off[:-1] + 7
            self._tshape = self._elem[shape_elem_id]
        return self._tshape

    @property
    def tshape_key(
        self, as_array=False
    ) -> Union[npt.NDArray[np.int32], Dict[int, npt.NDArray[np.int32]]]:
        """Return a dictionary with the mapping between element type and element shape.

        TShape is only applicable to contact elements.
        """
        if self._tshape_key is None:
            self._tshape_key = np.unique(np.vstack((self.et_id, self.tshape)), axis=1).T

        if as_array:
            return self._tshape_key
        return {elem_id: tshape for elem_id, tshape in self._tshape_key}


def fix_missing_midside(cells, nodes, celltypes, offset, angles, nnum):
    """Add missing midside nodes to cells.

    ANSYS sometimes does not add midside nodes, and this is denoted in
    the element array with a ``0``.  When translated to VTK, this is
    saved as a ``-1``.  If this is not corrected, VTK will segfault.

    This function creates missing midside nodes for the quadratic
    elements.
    """
    # Check for missing midside nodes
    mask = cells == -1
    nnodes = nodes.shape[0]

    nextra = mask.sum()
    cells[mask] = np.arange(nnodes, nnodes + nextra)

    nodes_new = np.empty((nnodes + nextra, 3))
    nodes_new[:nnodes] = nodes
    nodes_new[nnodes:] = 0  # otherwise, segfault disaster

    # Set new midside nodes directly between their edge nodes
    temp_nodes = nodes_new.copy()
    _relaxmidside.reset_midside(cells, celltypes, offset, temp_nodes)

    # merge midside nodes
    unique_nodes, idx_a, idx_b = unique_rows(temp_nodes[nnodes:])

    # rewrite node numbers
    cells[mask] = idx_b + nnodes
    nextra = idx_a.shape[0]  # extra unique nodes
    nodes_new = nodes_new[: nnodes + nextra]
    nodes_new[nnodes:] = unique_nodes

    if angles is not None:
        new_angles = np.empty((nnodes + nextra, 3))
        new_angles[:nnodes] = angles
        new_angles[nnodes:] = 0
    else:
        new_angles = None

    # Add extra node numbers
    nnum_new = np.empty(nnodes + nextra)
    nnum_new[:nnodes] = nnum
    nnum_new[nnodes:] = -1
    return nodes_new, new_angles, nnum_new



"""Test ``mapdl_archive/archive.py``."""

import os
import pathlib
import sys

import numpy as np
import pytest
import pyvista as pv
from pyvista import CellType
from pyvista import examples as pyvista_examples
from pyvista.plotting import system_supports_plotting

import mapdl_archive
from mapdl_archive import Archive, _archive, archive, examples

# Windows issues Python 3.12
WIN_PY312 = sys.version_info.minor == 12 and os.name == "nt"
skip_plotting = pytest.mark.skipif(
    WIN_PY312 or not system_supports_plotting(), reason="Requires active X Server"
)

LINEAR_CELL_TYPES = [
    CellType.TETRA,
    CellType.PYRAMID,
    CellType.WEDGE,
    CellType.HEXAHEDRON,
]
QUADRATIC_CELL_TYPES = [
    CellType.QUADRATIC_TETRA,
    CellType.QUADRATIC_PYRAMID,
    CellType.QUADRATIC_WEDGE,
    CellType.QUADRATIC_HEXAHEDRON,
]

TEST_PATH = os.path.dirname(os.path.abspath(__file__))
TESTFILES_PATH = os.path.join(TEST_PATH, "test_data")
TESTFILES_PATH_PATHLIB = pathlib.Path(TESTFILES_PATH)
DAT_FILE = os.path.join(TESTFILES_PATH, "Panel_Transient.dat")


def compare_dicts_with_arrays(dict1, dict2):
    if dict1.keys() != dict2.keys():
        return False
    for key in dict1:
        if not np.array_equal(dict1[key], dict2[key]):
            return False
    return True


def proto_cmblock(array):
    """prototype cmblock code"""
    items = np.zeros_like(array)
    items[0] = array[0]

    c = 1
    in_list = False
    for i in range(array.size - 1):
        # check if part of a range
        if array[i + 1] - array[i] == 1:
            in_list = True
        elif array[i + 1] - array[i] > 1:
            if in_list:
                items[c] = -array[i]
                c += 1
                items[c] = array[i + 1]
                c += 1
            else:
                items[c] = array[i + 1]
                c += 1
            in_list = False

    # check if we've ended on a list
    # catch if last item is part of a list
    if items[c - 1] != abs(array[-1]):
        items[c] = -array[i + 1]
        c += 1

    return items[:c]


@pytest.fixture()
def pathlib_archive():
    filename = TESTFILES_PATH_PATHLIB / "ErnoRadiation.cdb"
    return Archive(filename)


@pytest.fixture()
def hex_archive():
    return Archive(examples.hexarchivefile)


@pytest.fixture(scope="module")
def all_solid_cells_archive():
    return Archive(os.path.join(TESTFILES_PATH, "all_solid_cells.cdb"))


@pytest.fixture(scope="module")
def all_solid_cells_archive_linear():
    return Archive(
        os.path.join(TESTFILES_PATH, "all_solid_cells.cdb"), force_linear=True
    )


@pytest.mark.parametrize(
    "array",
    (
        np.arange(1, 10, dtype=np.int32),
        np.array([1, 5, 10, 20, 40, 80], dtype=np.int32),
        np.array([1, 2, 3, 10, 20, 40, 51, 52, 53], dtype=np.int32),
        np.array([1, 2, 3, 10, 20, 40], dtype=np.int32),
        np.array([10, 20, 40, 50, 51, 52], dtype=np.int32),
    ),
)
def test_cython_cmblock(array):
    """Simply verify it's identical to the prototype python code"""
    assert np.allclose(proto_cmblock(array), _archive.cmblock_items_from_array(array))


def test_load_dat():
    arch = Archive(DAT_FILE, read_parameters=True)
    assert arch.n_node == 1263  # through inspection of the dat file
    assert arch.n_elem == 160  # through inspection of the dat file
    assert "Panelflattern" in arch.parameters["_wb_userfiles_dir"]


def test_repr(hex_archive):
    assert f"{hex_archive.n_node}" in str(hex_archive)
    assert f"{hex_archive.n_elem}" in str(hex_archive)


@skip_plotting
def test_plot(hex_archive):
    cpos = hex_archive.plot(return_cpos=True)
    assert isinstance(cpos, pv.CameraPosition)


def test_read_mesh200():
    archive = Archive(os.path.join(TESTFILES_PATH, "mesh200.cdb"))
    assert archive.grid.n_cells == 1000


def test_archive_init(hex_archive):
    assert isinstance(hex_archive._raw, dict)
    assert isinstance(hex_archive.grid, pv.UnstructuredGrid)

    assert hex_archive._nblock_start > 0
    assert hex_archive._nblock_end > hex_archive._nblock_start


def test_parse_vtk(hex_archive):
    grid = hex_archive.grid
    assert grid.points.size
    assert grid.cells.size
    assert "ansys_node_num" in grid.point_data

    with pytest.raises(TypeError):
        hex_archive._parse_vtk(allowable_types=-1)

    with pytest.raises(TypeError):
        hex_archive._parse_vtk(allowable_types=3.0)


def test_invalid_archive(tmpdir, hex_archive):
    nblock_filename = str(tmpdir.mkdir("tmpdir").join("nblock.cdb"))
    mapdl_archive.write_nblock(nblock_filename, hex_archive.nnum, hex_archive.nodes)

    archive = Archive(nblock_filename)
    assert archive.grid is None


def test_write_angle(tmpdir, hex_archive):
    nblock_filename = str(tmpdir.mkdir("tmpdir").join("nblock.cdb"))
    mapdl_archive.write_nblock(
        nblock_filename, hex_archive.nnum, hex_archive.nodes, hex_archive.node_angles
    )

    archive = Archive(nblock_filename, parse_vtk=False)
    assert np.allclose(archive.nodes, hex_archive.nodes)


def test_missing_midside():
    allowable_types = [45, 95, 185, 186, 92, 187]
    archive_file = os.path.join(TESTFILES_PATH, "mixed_missing_midside.cdb")
    archive = Archive(archive_file, allowable_types=allowable_types)

    assert not np.any(archive.grid.celltypes == CellType.TETRA)


def test_missing_midside_write(tmpdir):
    allowable_types = [45, 95, 185, 186, 92, 187]
    archive_file = os.path.join(TESTFILES_PATH, "mixed_missing_midside.cdb")
    archive = Archive(archive_file, allowable_types=allowable_types)

    filename = str(tmpdir.join("tmp.cdb"))
    with pytest.raises(RuntimeError, match="Unsupported element types"):
        mapdl_archive.save_as_archive(filename, archive.grid, exclude_missing=True)

    mapdl_archive.save_as_archive(
        filename, archive.grid, exclude_missing=True, reset_etype=True
    )
    archive_new = Archive(filename)

    with pytest.raises(TypeError, match="must be an UnstructuredGrid"):
        mapdl_archive.save_as_archive(filename, [1, 2, 3])


def test_writehex(tmpdir, hex_archive):
    filename = str(tmpdir.mkdir("tmpdir").join("tmp.cdb"))
    mapdl_archive.save_as_archive(filename, hex_archive.grid)
    archive_new = Archive(filename)
    assert np.allclose(hex_archive.grid.points, archive_new.grid.points)
    assert np.allclose(
        hex_archive.grid.cell_connectivity,
        archive_new.grid.cell_connectivity,
    )

    for node_component in hex_archive.node_components:
        assert np.allclose(
            hex_archive.node_components[node_component],
            archive_new.node_components[node_component],
        )

    for element_component in hex_archive.element_components:
        assert np.allclose(
            hex_archive.element_components[element_component],
            archive_new.element_components[element_component],
        )


def test_write_voxel(tmpdir):
    filename = str(tmpdir.join("tmp.cdb"))
    grid = pv.ImageData(dimensions=(10, 10, 10))
    mapdl_archive.save_as_archive(filename, grid)

    archive = Archive(filename)
    assert np.allclose(archive.grid.points, grid.points)
    assert np.allclose(archive.grid.point_data["ansys_node_num"], range(1, 1001))
    assert archive.grid.n_cells, grid.n_cells


def test_writesector(tmpdir):
    archive = Archive(examples.sector_archive_file)
    filename = str(tmpdir.mkdir("tmpdir").join("tmp.cdb"))
    mapdl_archive.save_as_archive(filename, archive.grid)
    archive_new = Archive(filename)

    assert np.allclose(archive.grid.points, archive_new.grid.points)
    assert np.allclose(archive.grid.cells, archive_new.grid.cells)


def test_writehex_missing_elem_num(tmpdir, hex_archive):
    grid = hex_archive.grid
    grid.cell_data["ansys_elem_num"][:10] = -1
    grid.cell_data["ansys_etype"] = np.ones(grid.number_of_cells) * -1
    grid.cell_data["ansys_elem_type_num"] = np.ones(grid.number_of_cells) * -1

    filename = str(tmpdir.mkdir("tmpdir").join("tmp.cdb"))
    mapdl_archive.save_as_archive(filename, grid)
    archive_new = Archive(filename)

    assert np.allclose(hex_archive.grid.points, archive_new.grid.points)
    assert np.allclose(hex_archive.grid.cells, archive_new.grid.cells)


def test_writehex_missing_node_num(tmpdir, hex_archive):
    hex_archive.grid.point_data["ansys_node_num"][:-1] = -1

    filename = str(tmpdir.mkdir("tmpdir").join("tmp.cdb"))
    mapdl_archive.save_as_archive(filename, hex_archive.grid)
    archive_new = Archive(filename)
    assert np.allclose(hex_archive.grid.points.shape, archive_new.grid.points.shape)
    assert np.allclose(hex_archive.grid.cells.size, archive_new.grid.cells.size)


def test_write_non_ansys_grid(tmpdir):
    grid = pv.UnstructuredGrid(pyvista_examples.hexbeamfile)
    del grid.point_data["sample_point_scalars"]
    del grid.cell_data["sample_cell_scalars"]
    archive_file = str(tmpdir.mkdir("tmpdir").join("tmp.cdb"))
    mapdl_archive.save_as_archive(archive_file, grid)


def test_read_complex_archive(all_solid_cells_archive):
    nblock_expected = np.array(
        [
            [3.7826539829200e00, 1.2788958692644e00, -1.0220880953640e00],
            [3.7987359490873e00, 1.2312085780780e00, -1.0001885444969e00],
            [3.8138798206653e00, 1.1833200772896e00, -9.7805743587145e-01],
            [3.7751258193793e00, 1.2956563072306e00, -9.9775569295981e-01],
            [3.7675976558386e00, 1.3124167451968e00, -9.7342329055565e-01],
            [3.8071756567432e00, 1.2018089624856e00, -9.5159140433025e-01],
            [3.8004714928212e00, 1.2202978476816e00, -9.2512537278904e-01],
            [3.7840345743299e00, 1.2663572964392e00, -9.4927433167235e-01],
            [3.8682501483615e00, 1.4211343558710e00, -9.2956245308371e-01],
            [3.8656154427804e00, 1.4283573726940e00, -9.3544082975315e-01],
            [3.8629807371994e00, 1.4355803895169e00, -9.4131920642259e-01],
            [3.8698134427618e00, 1.4168612083433e00, -9.3457292477788e-01],
            [3.8645201728196e00, 1.4314324609914e00, -9.4526873324423e-01],
            [3.8713767371621e00, 1.4125880608155e00, -9.3958339647206e-01],
            [3.8687181728010e00, 1.4199362966407e00, -9.4440082826897e-01],
            [3.8660596084399e00, 1.4272845324660e00, -9.4921826006588e-01],
            [3.7847463501820e00, 1.2869612289286e00, -1.0110875234148e00],
            [3.7882161293470e00, 1.2952473975570e00, -1.0006326084202e00],
            [3.7840036708439e00, 1.3089808408341e00, -9.8189659453120e-01],
            [3.7736944340897e00, 1.3175655146540e00, -9.6829193559890e-01],
            [3.7797912123408e00, 1.3227142841112e00, -9.6316058064216e-01],
            [3.8163322819008e00, 1.1913589544053e00, -9.6740419078720e-01],
            [3.8046827481496e00, 1.2474593204382e00, -9.7922600135387e-01],
            [3.8202228218151e00, 1.1995824283636e00, -9.5733187068101e-01],
            [3.9797161316330e00, 2.5147820926190e-01, -5.1500799817626e-01],
            [3.9831382922541e00, 2.0190980565891e-01, -5.0185526897444e-01],
            [3.9810868976408e00, 2.3910377061737e-01, -5.4962360790281e-01],
            [3.9772930845240e00, 2.8865001362748e-01, -5.6276585706615e-01],
            [3.9816265976187e00, 2.1428739259987e-01, -4.6723916677654e-01],
            [3.9839413943097e00, 1.8949722823843e-01, -5.3648152416530e-01],
            [3.7962006776348e00, 1.2764624207283e00, -9.3931008487698e-01],
            [3.8126101429289e00, 1.2302105573453e00, -9.1545958911180e-01],
            [3.8065408178751e00, 1.2252542025135e00, -9.2029248095042e-01],
            [3.8164164823720e00, 1.2148964928545e00, -9.3639572989640e-01],
            [3.8972892823450e00, 2.7547119775919e-01, -5.6510422311694e-01],
            [3.9015993648189e00, 2.0235606714652e-01, -4.6987255385930e-01],
            [3.9023812010290e00, 1.7705558022279e-01, -5.3881795411458e-01],
            [3.9019902829240e00, 1.8970582368465e-01, -5.0434525398694e-01],
            [3.8998352416870e00, 2.2626338899099e-01, -5.5196108861576e-01],
            [3.8994443235820e00, 2.3891363245285e-01, -5.1748838848812e-01],
            [3.9372911834345e00, 2.8206060569333e-01, -5.6393504009155e-01],
            [3.9416129812188e00, 2.0832172987319e-01, -4.6855586031792e-01],
            [3.9431612976694e00, 1.8327640423061e-01, -5.3764973913994e-01],
            [3.8619577233846e00, 1.4192189812407e00, -9.2587403626770e-01],
            [3.8507167163959e00, 1.4238788373222e00, -9.3661710728291e-01],
            [3.8651039358730e00, 1.4201766685559e00, -9.2771824467570e-01],
            [3.8624692302920e00, 1.4273996853788e00, -9.3359662134515e-01],
            [3.8610467267790e00, 1.4182334490688e00, -9.3810025187748e-01],
            [3.8563372198902e00, 1.4215489092814e00, -9.3124557177530e-01],
            [3.8568487267976e00, 1.4297296134196e00, -9.3896815685275e-01],
            [3.8583881624179e00, 1.4255816848941e00, -9.4291768367439e-01],
            [3.8594834323787e00, 1.4225065965966e00, -9.3308978018331e-01],
        ]
    )
    assert np.allclose(nblock_expected, all_solid_cells_archive.nodes)

    grid = all_solid_cells_archive.grid
    assert grid.number_of_cells == 4
    assert np.unique(grid.celltypes).size == 4
    assert np.all(grid.celltypes > 20)


def test_read_complex_archive_linear(all_solid_cells_archive_linear):
    grid = all_solid_cells_archive_linear.grid
    assert np.all(grid.celltypes < 20)


@pytest.mark.parametrize("celltype", QUADRATIC_CELL_TYPES)
def test_write_quad_complex_archive(tmpdir, celltype, all_solid_cells_archive):
    grid = all_solid_cells_archive.grid
    mask = grid.celltypes == celltype
    assert mask.any()
    grid = grid.extract_cells(mask)

    tmp_archive_file = str(tmpdir.mkdir("tmpdir").join("tmp.cdb"))

    mapdl_archive.save_as_archive(tmp_archive_file, grid)
    new_archive = Archive(tmp_archive_file)
    assert np.allclose(grid.cells, new_archive.grid.cells)
    assert np.allclose(grid.points, new_archive.grid.points)


@pytest.mark.parametrize("celltype", LINEAR_CELL_TYPES)
def test_write_lin_archive(tmpdir, celltype, all_solid_cells_archive_linear):
    linear_grid = all_solid_cells_archive_linear.grid

    mask = linear_grid.celltypes == celltype
    assert mask.any()
    linear_grid = linear_grid.extract_cells(mask)

    tmp_archive_file = str(tmpdir.mkdir("tmpdir").join("tmp.cdb"))

    mapdl_archive.save_as_archive(tmp_archive_file, linear_grid)
    new_archive = Archive(tmp_archive_file)
    assert np.allclose(linear_grid.celltypes, new_archive.grid.celltypes)


@pytest.mark.parametrize("as_array", [False, True])
def test_write_component(tmpdir, as_array):
    items = [1, 20, 50, 51, 52, 53]
    if as_array:
        items = np.asarray(items)
    filename = str(tmpdir.mkdir("tmpdir").join("tmp.cdb"))

    comp_name = "TEST"
    mapdl_archive.write_cmblock(filename, items, comp_name, "node")
    archive = Archive(filename)
    assert np.allclose(archive.node_components[comp_name], items)


def test_write_component_edge_case(tmpdir):
    items = np.arange(2, 34, step=2)
    filename = str(tmpdir.mkdir("tmpdir").join("tmp.cdb"))

    comp_name = "TEST"
    mapdl_archive.write_cmblock(filename, items, comp_name, "node")
    archive = Archive(filename)
    assert np.allclose(archive.node_components[comp_name], items)

    with pytest.raises(ValueError, match="must be either"):
        mapdl_archive.write_cmblock(filename, items, comp_name, "FOO")


def test_read_parm():
    filename = os.path.join(TESTFILES_PATH, "parm.cdb")
    archive = Archive(filename)
    with pytest.raises(AttributeError):
        archive.parameters

    archive = Archive(filename, read_parameters=True)
    assert len(archive.parameters) == 2
    for parm in archive.parameters:
        assert isinstance(archive.parameters[parm], np.ndarray)


def test_read_wb_nblock():
    expected = np.array(
        [
            [9.89367578e-02, -8.07092192e-04, 8.53764953e00],
            [9.65803244e-02, 2.00906704e-02, 8.53744951e00],
            [9.19243555e-02, 3.98781615e-02, 8.53723652e00],
        ]
    )
    filename = os.path.join(TESTFILES_PATH, "workbench_193.cdb")
    archive = Archive(filename)
    assert np.allclose(archive.nodes, expected)
    assert np.allclose(archive.node_angles, 0)


def test_read_hypermesh():
    expected = np.array(
        [
            [-6.01203, 2.98129, 2.38556],
            [-3.03231, 2.98067, 2.38309],
            [-0.03485, 2.98004, 2.3805],
            [2.98794, 2.97941, 2.37773],
            [5.98956, 2.97878, 2.37488],
            [5.98956, 5.97878, 2.37488],
        ]
    )

    filename = os.path.join(TESTFILES_PATH, "hypermesh.cdb")
    archive = Archive(filename, verbose=True)
    assert np.allclose(archive.nodes[:6], expected)


@pytest.mark.parametrize("angles", [True, False])
def test_cython_write_nblock(hex_archive, tmpdir, angles):
    nblock_filename = str(tmpdir.mkdir("tmpdir").join("nblock.inp"))

    if angles:
        _archive.py_write_nblock(
            nblock_filename,
            hex_archive.nnum,
            hex_archive.nnum[-1],
            hex_archive.nodes,
            hex_archive.node_angles,
        )
    else:
        _archive.py_write_nblock(
            nblock_filename,
            hex_archive.nnum,
            hex_archive.nnum[-1],
            hex_archive.nodes,
            np.empty((0, 0)),
        )

    tmp_archive = Archive(nblock_filename)
    assert np.allclose(hex_archive.nnum, tmp_archive.nnum)
    assert np.allclose(hex_archive.nodes, tmp_archive.nodes)
    if angles:
        assert np.allclose(hex_archive.node_angles, tmp_archive.node_angles)


@pytest.mark.parametrize("has_angles", [True, False])
@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_write_nblock(hex_archive, tmpdir, dtype, has_angles):
    nblock_filename = str(tmpdir.mkdir("tmpdir").join("nblock.inp"))

    nodes = hex_archive.nodes.astype(dtype)
    if has_angles:
        angles = hex_archive.node_angles
    else:
        angles = None
    archive.write_nblock(nblock_filename, hex_archive.nnum, nodes, angles, mode="w")

    tmp_archive = Archive(nblock_filename)
    assert np.allclose(hex_archive.nnum, tmp_archive.nnum)
    assert np.allclose(hex_archive.nodes, tmp_archive.nodes)
    if has_angles:
        assert np.allclose(hex_archive.node_angles, tmp_archive.node_angles)


@pytest.mark.parametrize("sig_digits", [8, 18])
def test_write_nblock_sig_digits(hex_archive, tmpdir, sig_digits):
    nblock_filename = str(tmpdir.mkdir("tmpdir").join("nblock.inp"))

    nodes = hex_archive.nodes
    angles = hex_archive.node_angles
    with pytest.raises(ValueError, match="sig_digits"):
        archive.write_nblock(
            nblock_filename, hex_archive.nnum, nodes, angles, sig_digits=-1
        )

    archive.write_nblock(
        nblock_filename, hex_archive.nnum, nodes, angles, sig_digits=sig_digits
    )  # more than 18 fails
    tmp_archive = Archive(nblock_filename)
    assert np.allclose(hex_archive.nnum, tmp_archive.nnum)
    assert np.allclose(hex_archive.nodes, tmp_archive.nodes)
    assert np.allclose(hex_archive.node_angles, tmp_archive.node_angles)


def test_cython_write_eblock(hex_archive, tmpdir):
    filename = str(tmpdir.mkdir("tmpdir").join("eblock.inp"))

    etype = np.ones(hex_archive.n_elem, np.int32)
    typenum = hex_archive.etype
    elem_nnodes = np.empty(etype.size, np.int32)
    elem_nnodes[typenum == 181] = 4
    elem_nnodes[typenum == 185] = 8
    elem_nnodes[typenum == 186] = 20
    elem_nnodes[typenum == 187] = 10
    nodenum = hex_archive.nnum

    cells, offset = hex_archive.grid.cell_connectivity, hex_archive.grid.offset
    _archive.py_write_eblock(
        filename,
        hex_archive.enum,
        etype,
        hex_archive.material_type,
        np.ones(hex_archive.n_elem, np.int32),
        elem_nnodes,
        cells.astype(np.int32, copy=False),
        offset.astype(np.int32, copy=False),
        hex_archive.grid.celltypes,
        typenum,
        nodenum,
    )


def test_rlblock_prior_to_nblock():
    # test edge case where RLBLOCK is immediately prior to the NBLOCK
    filename = os.path.join(TESTFILES_PATH, "ErnoRadiation.cdb")
    archive = Archive(filename)
    assert archive.n_node == 65
    assert archive.n_elem == 36


def test_etblock():
    # test edge case where RLBLOCK is immediately prior to the NBLOCK
    filename = os.path.join(TESTFILES_PATH, "etblock.cdb")
    archive = Archive(filename)
    assert archive.n_node == 4
    assert archive.n_elem == 1


def test_overwrite_nblock(tmpdir, hex_archive):
    # ensure that we capture the entire NBLOCK
    with open(hex_archive._filename, "rb") as fid:
        fid.seek(hex_archive._nblock_start)
        nblock_txt = fid.read(
            hex_archive._nblock_end - hex_archive._nblock_start
        ).decode()

    assert nblock_txt.startswith("NBLOCK")
    assert nblock_txt.splitlines()[-1].endswith("-1,")

    filename = str(tmpdir.mkdir("tmpdir").join("tmp.cdb"))
    nodes = np.random.random(hex_archive.nodes.shape)
    hex_archive.overwrite_nblock(
        filename,
        hex_archive.nnum,
        nodes,
    )

    archive_new = Archive(filename)
    assert np.allclose(nodes, archive_new.grid.points)
    assert np.allclose(hex_archive.nnum, archive_new.nnum)
    assert np.allclose(hex_archive.grid.cells.size, archive_new.grid.cells.size)


class TestPathlibFilename:
    def test_pathlib_filename_property(self, pathlib_archive):
        assert isinstance(pathlib_archive.pathlib_filename, pathlib.Path)

    def test_filename_property_is_string(self, pathlib_archive):
        filename = TESTFILES_PATH_PATHLIB / "ErnoRadiation.cdb"
        a = Archive(filename)
        assert isinstance(a.filename, str)

    def test_filename_setter_pathlib(self, pathlib_archive):
        with pytest.raises(AttributeError):
            pathlib_archive.filename = pathlib.Path("dummy2")

    def test_filename_setter_string(self, pathlib_archive):
        with pytest.raises(AttributeError):
            pathlib_archive.filename = "dummy2"


def test_save_as_numpy(tmpdir, hex_archive):
    """Test saving to and loading from a npz file."""
    path = str(tmpdir.join("data.npz"))
    hex_archive.save_as_numpy(path)
    hex_in = Archive(path)

    assert hex_archive._raw.keys() == hex_in._raw.keys()

    for key, value in hex_archive._raw.items():
        if isinstance(value, np.ndarray):
            assert np.allclose(value, hex_in._raw[key])
        else:
            if isinstance(value, dict):
                compare_dicts_with_arrays(value, hex_in._raw[key])
            else:
                assert value == hex_in._raw[key]



# cython: language_level=3
# cython: infer_types=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: cdivision=True
# cython: nonecheck=False
# cython: embedsignature=True

from libc.math cimport abs
from libc.stdio cimport FILE, fclose, fdopen, fopen, fwrite

import numpy as np

cimport numpy as np

ctypedef unsigned char uint8_t

cdef extern from 'archive.h' nogil:
    int write_nblock(FILE*, const int, const int, const int*, const double*,
                     const double*, int, int)
    int write_nblock_float(FILE*, const int, const int, const int*, const float*,
                           const float*, int, int)
    int write_eblock(FILE*, const int, const int*, const int*, const int*,
                     const int*, const int*, const uint8_t*, const int*,
                     const int*, const int*, const int*);


cdef extern from "stdio.h":
    FILE *fdopen(int, const char *)



def py_write_nblock(filename, const int [::1] node_id, int max_node_id,
                    const double [:, ::1] pos, const double [:, ::1] angles,
                    mode='w', sig_digits=13):
    """Write a node block to a file.

    Parameters
    ----------
    filename : str
        Name of the file to write to.

    node_id : np.ndarray
        Array of node ids.

    max_node_id : int
        Maximum node id.

    pos : np.ndarray[double]
        2D array of node coordinates. Each row represents a node, and the
        columns are the x, y, and z coordinates.

    angles : np.ndarray[double], optional
        2D array of angles. Each row represents a node, and the columns are the
        angles. If not provided, all angles will be set to zero.

    mode : str, optional
        File open mode. Default is 'w' (write), other options include 'a' (append).

    sig_digits : int, default: 13
        Number of significant digits to use when writing the nodes.

    Raises
    ------
    IOError
        If the file cannot be opened for writing.

    Notes
    -----
    The file will be written in the ANSYS node block format. Each line of the
    file will contain the id, coordinates, and (optionally) angles for a single
    node.

    """
    # attach the stream to the python file
    cdef FILE* cfile = fopen(filename.encode(), mode.encode())

    cdef int n_nodes = pos.shape[0]

    cdef int has_angles = 0
    if angles.size == pos.size:
        has_angles = 1
    else:
        angles = np.zeros((1, 1), np.double)
    write_nblock(cfile, n_nodes, max_node_id, &node_id[0], &pos[0, 0],
                 &angles[0, 0], has_angles, sig_digits);
    fclose(cfile)


def py_write_nblock_float(filename, const int [::1] node_id, int max_node_id,
                          const float [:, ::1] pos, const float [:, ::1] angles,
                          mode='w', sig_digits=13):
    """Write a float32 node block to a file.

    Parameters
    ----------
    filename : str
        Name of the file to write to.

    node_id : np.ndarray
        Array of node ids.

    max_node_id : int
        Maximum node id.

    pos : np.ndarray[double]
        2D array of node coordinates. Each row represents a node, and the
        columns are the x, y, and z coordinates.

    angles : np.ndarray[double], optional
        2D array of angles. Each row represents a node, and the columns are the
        angles. If not provided, all angles will be set to zero.

    mode : str, optional
        File open mode. Default is 'w' (write), other options include 'a' (append).

    sig_digits : int, default: 13
        Number of significant digits to use when writing the nodes.

    Raises
    ------
    IOError
        If the file cannot be opened for writing.

    Notes
    -----
    The file will be written in the ANSYS node block format. Each line of the
    file will contain the id, coordinates, and (optionally) angles for a single
    node.

    """
    # attach the stream to the python file
    cdef FILE* cfile = fopen(filename.encode(), mode.encode())

    cdef int n_nodes = pos.shape[0]
    cdef int has_angles = 0
    if angles.size == pos.size:
        has_angles = 1
    else:
        angles = np.zeros((1, 1), np.float32)
    write_nblock_float(cfile, n_nodes, max_node_id, &node_id[0], &pos[0, 0],
                       &angles[0, 0], has_angles, sig_digits);
    fclose(cfile)


def py_write_eblock(
    filename,
    const int [::1] elem_id,
    const int [::1] etype,
    const int [::1] mtype,
    const int [::1] rcon,
    const int [::1] elem_nnodes,
    const int [::1] cells,
    const int [::1] offset,
    const uint8_t [::1] celltypes,
    const int [::1] typenum,
    const int [::1] nodenum,
    mode='w'):
    cdef FILE* cfile = fopen(filename.encode(), mode.encode())
    write_eblock(
        cfile,
        celltypes.size,
        &elem_id[0],
        &etype[0],
        &mtype[0],
        &rcon[0],
        &elem_nnodes[0],
        &celltypes[0],
        &offset[0],
        &cells[0],
        &typenum[0],
        &nodenum[0]
    )
    fclose(cfile)


def cmblock_items_from_array(int [::1] array):
    """Given a list of items, convert to a ANSYS formatted CMBLOCK.

    For example ``1, 2, 3, 4, 8``

    will be converted to

    ``1, -4, 8``

    Where the -4 indicates all the items between 1 and -4.
    """

    # first, verify items in array are sorted
    cdef int i
    cdef is_sorted = 1
    for i in range(array.size - 1):
        if array[i] > array[i + 1]:
            is_sorted = 0
            break

    if not is_sorted:
        array = np.unique(array)

    cdef int [::1] items = np.empty_like(array)
    items[0] = array[0]

    cdef int c = 1
    cdef int in_list = 0
    for i in range(array.size - 1):
        # check if part of a range
        if array[i + 1] - array[i] == 1:
            in_list = 1
        elif array[i + 1] - array[i] > 1:
            if in_list:
                items[c] = -array[i]; c += 1
                items[c] = array[i + 1]; c += 1
            else:
                items[c] = array[i + 1]; c += 1
            in_list = 0

    # catch if last item is part of a list
    if items[c - 1] != abs(array[array.size - 1]):
        items[c] = -array[array.size - 1]; c += 1

    return np.array(items[:c])



"""Module to read MAPDL ASCII block formatted CDB files."""

import io
import logging
import os
import pathlib
import shutil
from typing import Dict, List, Optional, Sequence, Union, cast

import numpy as np
import numpy.typing as npt
from pyvista import CellType, UnstructuredGrid

from mapdl_archive import _archive, _reader
from mapdl_archive.mesh import Mesh

# types
NPArray_FLOAT32_64 = npt.NDArray[Union[np.float32, np.float64]]
NPArray_FLOAT32 = npt.NDArray[np.float32]
NPArray_FLOAT64 = npt.NDArray[np.float64]

VTK_VOXEL = 11

log = logging.getLogger(__name__)
log.setLevel("CRITICAL")


class Archive(Mesh):
    """Read a blocked MAPDL archive file or input file.

    Reads a blocked CDB file and optionally parses it to a vtk grid.
    This can be used to read in files written from MAPDL using the
    ``CDWRITE`` command or input files (``'.dat'``) files written from
    ANSYS Workbench.

    Write the archive file using ``CDWRITE, DB, archive.cdb``

    Parameters
    ----------
    filename : string, pathlib.Path
        Supports blocked MAPDL archive .cdb files or .npz written from this
        class using :class:`Archive.save_as_numpy`.

    read_parameters : bool, default: False
        Optionally read parameters from the archive file.

    parse_vtk : bool, default: True
        When ``True``, parse the raw data into to VTK format.

    force_linear : bool, default: False
        This parser creates quadratic elements if available. Set
        this to ``True`` to always create linear elements.

    allowable_types : list, optional
        Allowable element types.  Defaults to all valid element
        types in ``mapdl_archive.elements.valid_types``

        See ``help(mapdl_archive.elements)`` for available element types.

    null_unallowed : bool, default: False
        Elements types not matching element types will be stored
        as empty (null) elements.  Useful for debug or tracking
        element numbers.

    verbose : bool, optional
        Print out each step when reading the archive file.  Used for
        debug purposes and defaults to ``False``.

    name : str, optional
        Internally used parameter used to have a custom ``__repr__``.

    read_eblock : bool, default: True
        Read the element block.

    Examples
    --------
    >>> import mapdl_archive
    >>> from mapdl_archive import examples
    >>> hex_beam = mapdl_archive.Archive(examples.hexarchivefile)
    >>> print(hex_beam)
    ANSYS Archive File HexBeam.cdb
      Number of Nodes:              40
      Number of Elements:           321
      Number of Element Types:      1
      Number of Node Components:    2
      Number of Element Components: 2

    Print the node array.

    >>> hex_beam.nodes
    array([[0.  , 0.  , 0.  ],
           [1.  , 0.  , 0.  ],
           [0.25, 0.  , 0.  ],
           ...,
           [0.75, 0.5 , 3.5 ],
           [0.75, 0.5 , 4.  ],
           [0.75, 0.5 , 4.5 ]])

    Read an Ansys mechanical input file.

    >>> my_archive = mapdl_archive.Archive("C:/Users/user/ds.dat")

    Notes
    -----
    This class only reads EBLOCK records with SOLID records.  For
    example, the record ``EBLOCK,19,SOLID,,3588`` will be read, but
    ``EBLOCK,10,,,3588`` will not be read. Generally, MAPDL will only
    write SOLID records and Mechanical Workbench may write SOLID
    records. These additional records will be ignored.
    """

    def __init__(
        self,
        filename: Union[str, pathlib.Path],
        read_parameters: bool = False,
        parse_vtk: bool = True,
        force_linear: bool = False,
        allowable_types: Optional[List[Union[str, int]]] = None,
        null_unallowed: bool = False,
        verbose: bool = False,
        name: str = "",
        read_eblock: bool = True,
    ):
        """Initialize an instance of the archive class."""
        self._read_parameters: bool = read_parameters
        self._filename: pathlib.Path = pathlib.Path(filename)
        self._name: str = name
        if self.filename.endswith("npz"):
            npz_file = np.load(self.filename, allow_pickle=True)

            self._raw: _reader.ReadReturnDict = {
                "rnum": npz_file["rnum"],
                "rdat": npz_file["rdat"].tolist(),
                "ekey": npz_file["ekey"],
                "nnum": npz_file["nnum"],
                "nodes": npz_file["nodes"],
                "elem": npz_file["elem"],
                "elem_off": npz_file["elem_off"],
                "node_comps": npz_file["node_comps"].item(),
                "elem_comps": npz_file["elem_comps"].item(),
                "keyopt": npz_file["keyopt"].item(),
                "parameters": npz_file["parameters"].item(),
                "nblock_start": npz_file["nblock_start"],
                "nblock_end": npz_file["nblock_end"],
            }
        else:
            self._raw = _reader.read(
                self.filename,
                read_parameters=read_parameters,
                debug=verbose,
                read_eblock=read_eblock,
            )
        super().__init__(
            self._raw["nnum"],
            self._raw["nodes"],
            self._raw["elem"],
            self._raw["elem_off"],
            self._raw["ekey"],
            self._raw["rnum"],
            node_comps=self._raw["node_comps"],
            elem_comps=self._raw["elem_comps"],
            rdat=self._raw["rdat"],
            keyopt=self._raw["keyopt"],
        )

        self._allowable_types: Optional[list[Union[str, int]]] = allowable_types
        self._force_linear: bool = force_linear
        self._null_unallowed: bool = null_unallowed

        if parse_vtk:
            self._grid = self._parse_vtk(allowable_types, force_linear, null_unallowed)

    @property
    def filename(self) -> str:
        """Return the filename."""
        return str(self._filename)

    @property
    def pathlib_filename(self) -> pathlib.Path:
        """Return the filename as a ``pathlib.Path``."""
        return self._filename

    @property
    def parameters(self) -> Dict[str, np.ndarray]:
        """Return the parameters stored in the archive file.

        Examples
        --------
        >>> import mapdl_archive
        >>> from mapdl_archive import examples
        >>> archive = mapdl_archive.Archive(
        ...     examples.hexarchivefile, read_parameters=True
        ... )
        >>> archive.parameters
        {}

        """
        if not self._read_parameters:
            raise AttributeError(
                "No parameters read.  Read the archive again "
                " with ``read_parameters=True``"
            )
        return self._raw["parameters"]

    def __repr__(self) -> str:
        """Return the representation of the archive."""
        if self._name:
            txt: str = f"MAPDL {self._name}\n"
        else:
            basename = os.path.basename(self._filename)
            txt = f"ANSYS Archive File {basename}\n"

        txt += f"  Number of Nodes:              {len(self.nnum)}\n"
        txt += f"  Number of Elements:           {len(self.enum)}\n"
        txt += f"  Number of Element Types:      {len(self.ekey)}\n"
        txt += f"  Number of Node Components:    {len(self.node_components)}\n"
        txt += f"  Number of Element Components: {len(self.element_components)}\n"
        return txt

    @property
    def grid(self) -> UnstructuredGrid:
        """Return a ``pyvista.UnstructuredGrid`` of the archive file.

        Examples
        --------
        >>> import mapdl_archive
        >>> from mapdl_archive import examples
        >>> archive = mapdl_archive.Archive(examples.hexarchivefile)
        >>> archive.grid
        UnstructuredGrid (0x7ffa237f08a0)
          N Cells:      40
          N Points:     321
          X Bounds:     0.000e+00, 1.000e+00
          Y Bounds:     0.000e+00, 1.000e+00
          Z Bounds:     0.000e+00, 5.000e+00
          N Arrays:     13
        """
        if self._grid is None:  # parse the grid using the cached parameters
            self._grid = self._parse_vtk(
                self._allowable_types, self._force_linear, self._null_unallowed
            )
        return self._grid

    def plot(self, *args, **kwargs):
        """Plot the mesh.

        See ``help(pyvista.plot)`` for all optional kwargs.

        """
        if self._grid is None:  # pragma: no cover
            raise AttributeError(
                "Archive must be parsed as a vtk grid.\n Set `parse_vtk=True`"
            )
        kwargs.setdefault("color", "w")
        kwargs.setdefault("show_edges", True)
        return self.grid.plot(*args, **kwargs)

    @property
    def _nblock_start(self) -> int:
        """Return the start of the node block in the original file."""
        return self._raw["nblock_start"]

    @property
    def _nblock_end(self) -> int:
        """Return the end of the node block in the original file."""
        return self._raw["nblock_end"]

    def overwrite_nblock(self, filename, node_id, pos, angles=None, sig_digits=13):
        """Write out an archive file to disk while replacing its NBLOCK.

        Parameters
        ----------
        filename : str or file handle
            Filename to write node block to.
        node_id : numpy.ndarray
            ANSYS node numbers.
        pos : np.ndarray
            Array of node coordinates.
        angles : numpy.ndarray, optional
            Writes the node angles for each node when included.
        sig_digits : int, default: 13
            Number of significant digits to use when writing the nodes. Must be
            greater than 0.

        Examples
        --------
        Write a new archive file that overwrites the NBLOCK with random nodes
        and reuse the existing node numbers.

        >>> import numpy as np
        >>> import mapdl_archive
        >>> archive = mapdl_archive.Archive(examples.hexarchivefile)
        >>> new_nodes = np.random.random(archive.nodes.shape)
        >>> archive.overwrite_nblock("new_archive.cdb", archive.nnum, new_nodes)

        """
        with open(self._filename, "rb") as src_file, open(filename, "wb") as dest_file:
            # Copy the beginning of the file up to _nblock_start
            dest_file.write(src_file.read(self._nblock_start))

        # Write new nblock
        write_nblock(
            filename, node_id, pos, angles=angles, mode="a", sig_digits=sig_digits
        )

        # Copy the rest of the original file
        with open(self._filename, "rb") as src_file, open(filename, "ab") as dest_file:
            src_file.seek(self._nblock_end)
            dest_file.seek(0, io.SEEK_END)

            shutil.copyfileobj(src_file, dest_file)

    def save_as_numpy(self, filename: str):
        """Save this archive as a numpy "npz" file.

        This reduces the file size by around 50% compared with the Ansys
        blocked file format.

        """
        if not filename.endswith("npz"):
            raise ValueError("Filename should end with 'npz'")

        np.savez(filename, **self._raw)  # type: ignore


def save_as_archive(
    filename: Union[pathlib.Path, str],
    grid: UnstructuredGrid,
    mtype_start: int = 1,
    etype_start: int = 1,
    real_constant_start: int = 1,
    mode: str = "w",
    enum_start: int = 1,
    nnum_start: int = 1,
    include_etype_header: bool = True,
    reset_etype: bool = False,
    allow_missing: bool = True,
    include_surface_elements: bool = True,
    include_solid_elements: bool = True,
    include_components: bool = True,
    exclude_missing: bool = False,
    node_sig_digits: int = 13,
) -> None:
    """Write FEM as an ANSYS APDL archive file.

    This function supports the following element types:

        - ``vtk.VTK_HEXAHEDRON``
        - ``vtk.VTK_PYRAMID``
        - ``vtk.VTK_QUADRATIC_HEXAHEDRON``
        - ``vtk.VTK_QUADRATIC_PYRAMID``
        - ``vtk.VTK_QUADRATIC_TETRA``
        - ``vtk.VTK_QUADRATIC_WEDGE``
        - ``vtk.VTK_QUAD``
        - ``vtk.VTK_TETRA``
        - ``vtk.VTK_TRIANGLE``
        - ``vtk.VTK_VOXEL``
        - ``vtk.VTK_WEDGE``

    Will automatically renumber nodes and elements if the FEM does not
    contain ANSYS node or element numbers.  Node numbers are stored as
    a point array ``"ansys_node_num"``, and cell numbers are stored as
    cell array ``"ansys_elem_num"``.

    Parameters
    ----------
    filename : str, pathlib.Path
       Filename to write archive file.

    grid : pyvista.DataSet
        Any :class:`pyvista.DataSet` that can be cast to a
        :class:`pyvista.UnstructuredGrid`.

    mtype_start : int, optional
        Material number to assign to elements.  Can be set manually by
        adding the cell array "mtype" to the unstructured grid.

    etype_start : int, optional
        Starting element type number.  Can be manually set by adding
        the cell array "ansys_etype" to the unstructured grid.

    real_constant_start : int, optional
        Starting real constant to assign to unset cells.  Can be
        manually set by adding the cell array "ansys_real_constant" to
        the unstructured grid.

    mode : str, optional
        File mode.  See ``help(open)``

    enum_start : int, optional
        Starting element number to assign to unset cells.  Can be
        manually set by adding the cell array "ansys_elem_num" to the
        unstructured grid.

    nnum_start : int, optional
        Starting element number to assign to unset points.  Can be
        manually set by adding the point array "ansys_node_num" to the
        unstructured grid.

    include_etype_header : bool, optional
        For each element type, includes element type command
        (e.g. "ET, 1, 186") in the archive file.

    reset_etype : bool, optional
        Resets element type.  Element types will automatically be
        determined by the shape of the element (i.e. quadradic
        tetrahedrals will be saved as SOLID187, linear hexahedrals as
        SOLID185).  Default True.

    include_surface_elements : bool, optional
        Includes surface elements when writing the archive file and
        saves them as SHELL181.

    include_solid_elements : bool, optional
        Includes solid elements when writing the archive file and
        saves them as SOLID185, SOLID186, or SOLID187.

    include_components : bool, optional
        Writes note components to file.  Node components must be
        stored within the unstructured grid as uint8 or bool arrays.

    exclude_missing : bool, default: False
        When ``allow_missing=True``, write ``0`` instead of renumbering
        nodes. This allows you to exclude midside nodes for certain element
        types (e.g. ``SOLID186``). Missing midside nodes are identified as
        ``-1`` in the ``"ansys_node_num"`` array.

    node_sig_digits : int, default: 13
        Number of significant digits to use when writing the nodes. Must be
        greater than 0.

    Examples
    --------
    Write a ``pyvista.UnstructuredGrid`` to ``"archive.cdb"``.

    >>> import mapdl_archive
    >>> from pyvista import examples
    >>> grid = examples.load_hexbeam()
    >>> pymapdl_reader.save_as_archive("archive.cdb", grid)

    """
    if hasattr(grid, "cast_to_unstructured_grid"):
        grid = grid.cast_to_unstructured_grid()

    if not isinstance(grid, UnstructuredGrid):
        raise TypeError(
            f"``grid`` argument must be an UnstructuredGrid, not {type(grid)}"
        )

    allowable = []
    if include_solid_elements:
        allowable.extend(
            [
                CellType.VOXEL,
                CellType.TETRA,
                CellType.QUADRATIC_TETRA,
                CellType.PYRAMID,
                CellType.QUADRATIC_PYRAMID,
                CellType.WEDGE,
                CellType.QUADRATIC_WEDGE,
                CellType.HEXAHEDRON,
                CellType.QUADRATIC_HEXAHEDRON,
            ]
        )

    if include_surface_elements:
        allowable.extend([CellType.TRIANGLE, CellType.QUAD])
        # VTK_QUADRATIC_TRIANGLE,
        # VTK_QUADRATIC_QUAD

    # extract allowable cell types
    mask = np.in1d(grid.celltypes, allowable)
    if not mask.any():
        ucelltypes = np.unique(grid.celltypes)
        allowable.sort()
        raise RuntimeError(
            f"`grid` contains no allowable cell types. Contains types {ucelltypes} "
            f"and only {allowable} are allowed.\n\n"
            "See https://vtk.org/doc/nightly/html/vtkCellType_8h_source.html "
            "for more details."
        )
    grid = grid.extract_cells(mask)

    header = "/PREP7\n"

    # node numbers
    if "ansys_node_num" in grid.point_data:
        nodenum = grid.point_data["ansys_node_num"]
    else:
        log.info("No ANSYS node numbers set in input. Adding default range")
        nodenum = np.arange(1, grid.n_points + 1, dtype=np.int32)

    missing_mask = nodenum == -1
    if np.any(missing_mask):
        if not allow_missing:
            raise RuntimeError(
                'Missing node numbers.  Exiting due "allow_missing=False"'
            )
        elif exclude_missing:
            log.info("Excluding missing nodes from archive file.")
            nodenum = nodenum.copy()
            nodenum[missing_mask] = 0
        else:
            start_num = nodenum.max() + 1
            if nnum_start > start_num:
                start_num = nnum_start
            nadd = np.sum(nodenum == -1)
            end_num = start_num + nadd
            log.info(
                "FEM missing some node numbers.  Adding node numbering "
                "from %d to %d",
                start_num,
                end_num,
            )
            nodenum[missing_mask] = np.arange(start_num, end_num, dtype=np.int32)

    # element block
    ncells = grid.n_cells
    if "ansys_elem_num" in grid.cell_data:
        enum = grid.cell_data["ansys_elem_num"]
    else:
        if not allow_missing:
            raise RuntimeError(
                'Missing node numbers. Exiting due "allow_missing=False"'
            )
        log.info(
            "No ANSYS element numbers set in input. "
            "Adding default range starting from %d",
            enum_start,
        )
        enum = np.arange(1, ncells + 1, dtype=np.int32)

    if np.any(enum == -1):
        if not allow_missing:
            raise RuntimeError(
                '-1 encountered in "ansys_elem_num".\nExiting due "allow_missing=False"'
            )

        start_num = enum.max() + 1
        if enum_start > start_num:
            start_num = enum_start
        nadd = np.sum(enum == -1)
        end_num = start_num + nadd
        log.info(
            "FEM missing some cell numbers.  Adding numbering from %d to %d",
            start_num,
            end_num,
        )
        enum[enum == -1] = np.arange(start_num, end_num, dtype=np.int32)

    # material type
    if "ansys_material_type" in grid.cell_data:
        mtype = grid.cell_data["ansys_material_type"]
    else:
        log.info(
            "No ANSYS element numbers set in input.  "
            "Adding default range starting from %d",
            mtype_start,
        )
        mtype = np.arange(1, ncells + 1, dtype=np.int32)

    if np.any(mtype == -1):
        log.info("FEM missing some material type numbers.  Adding...")
        mtype[mtype == -1] = mtype_start

    # real constant
    if "ansys_real_constant" in grid.cell_data:
        rcon = grid.cell_data["ansys_real_constant"]
    else:
        log.info(
            "No ANSYS element numbers set in input.  "
            + "Adding default range starting from %d",
            real_constant_start,
        )
        rcon = np.arange(1, ncells + 1, dtype=np.int32)

    if np.any(rcon == -1):
        log.info("FEM missing some material type numbers.  Adding...")
        rcon[rcon == -1] = real_constant_start

    # element type
    invalid = False
    if "ansys_etype" in grid.cell_data and not reset_etype:
        missing = False
        typenum = grid.cell_data["ansys_elem_type_num"]
        etype = grid.cell_data["ansys_etype"]
        if np.any(etype == -1):
            log.warning("Some elements are missing element type numbers.")
            invalid = True

        if include_etype_header and not invalid:
            _, ind = np.unique(etype, return_index=True)
            for idx in ind:
                header += "ET, %d, %d\n" % (etype[idx], typenum[idx])
    else:
        missing = True

    # check if valid
    if not missing:
        mask = grid.celltypes < 20
        if np.any(grid.cell_data["ansys_elem_type_num"][mask] == 186):
            invalid = True
            log.warning("Invalid ANSYS element types.")

    if invalid or missing:
        if not allow_missing:
            raise RuntimeError(
                'Invalid or missing data in "ansys_elem_type_num"'
                ' or "ansys_etype".  Exiting due "allow_missing=False"'
            )

        log.info(
            "No ANSYS element type or invalid data input.  "
            + "Adding default range starting from %d" % etype_start
        )

        etype = np.empty(grid.n_cells, np.int32)

        # VTK to SOLID186 mapping
        # TETRA delegated to SOLID187
        etype_186 = etype_start
        etype_186_types = [
            CellType.QUADRATIC_HEXAHEDRON,
            CellType.QUADRATIC_WEDGE,
            CellType.QUADRATIC_PYRAMID,
        ]
        etype[np.isin(grid.celltypes, etype_186_types)] = etype_186

        etype_187 = etype_start + 1
        etype[grid.celltypes == CellType.QUADRATIC_TETRA] = etype_187

        # VTK to SOLID185 mapping
        etype_185 = etype_start + 2
        etype_185_types = [
            CellType.VOXEL,
            CellType.TETRA,
            CellType.HEXAHEDRON,
            CellType.WEDGE,
            CellType.PYRAMID,
        ]
        etype[np.isin(grid.celltypes, etype_185_types)] = etype_185

        # Surface elements
        etype_181 = etype_start + 3
        etype_181_types = [
            CellType.TRIANGLE,
            CellType.QUAD,
        ]
        etype[np.isin(grid.celltypes, etype_181_types)] = etype_181

        typenum = np.empty_like(etype)
        typenum[etype == etype_185] = 185
        typenum[etype == etype_186] = 186
        typenum[etype == etype_187] = 187
        typenum[etype == etype_181] = 181

        header += f"ET,{etype_185},185\n"
        header += f"ET,{etype_186},186\n"
        header += f"ET,{etype_187},187\n"
        header += f"ET,{etype_181},181\n"

    # number of nodes written per element
    elem_nnodes = np.empty(etype.size, np.int32)
    elem_nnodes[typenum == 181] = 4
    elem_nnodes[typenum == 185] = 8
    elem_nnodes[typenum == 186] = 20
    elem_nnodes[typenum == 187] = 10

    if not reset_etype:
        unsup = np.setdiff1d(typenum, [181, 185, 186, 187])
        if unsup.any():
            raise RuntimeError(
                f"Unsupported element types {unsup}. Either set ``reset_etype=True``"
                " or remove (or relabel) the unsupported element types."
            )

    # edge case where element types are unsupported

    # write the EBLOCK
    filename = str(filename)
    with open(filename, mode) as f:
        f.write(header)

    if exclude_missing:
        log.info("Excluding missing nodes from archive file.")
        write_nblock(
            filename,
            nodenum[~missing_mask],
            grid.points[~missing_mask],
            mode="a",
            sig_digits=node_sig_digits,
        )
    else:
        write_nblock(
            filename, nodenum, grid.points, mode="a", sig_digits=node_sig_digits
        )

    # write remainder of eblock
    _write_eblock(
        filename,
        enum,
        etype,
        mtype,
        rcon,
        elem_nnodes,
        grid.cell_connectivity,
        grid.offset,
        grid.celltypes,
        typenum,
        nodenum,
        mode="a",
    )

    if include_components:
        with open(filename, "a") as fid:
            # write node components
            for node_key in grid.point_data:
                arr = grid.point_data[node_key]
                if arr.dtype in [np.uint8, np.bool_]:
                    items = nodenum[arr.view(np.bool_)]
                    write_cmblock(fid, items, node_key, "NODE")

            # write element components
            for node_key in grid.cell_data:
                arr = grid.cell_data[node_key]
                if arr.dtype in [np.uint8, np.bool_]:
                    items = enum[arr.view(np.bool_)]
                    write_cmblock(fid, items, node_key, "ELEM")


def write_nblock(
    filename: str,
    node_id: npt.NDArray[np.int_],
    pos: NPArray_FLOAT32_64,
    angles: Optional[NPArray_FLOAT32_64] = None,
    mode: str = "w",
    sig_digits: int = 13,
) -> None:
    """Write nodes and node angles to file.

    Parameters
    ----------
    filename : str
        Filename to write node block to.
    node_id : numpy.ndarray
        ANSYS node numbers.
    pos : np.ndarray
        Array of node coordinates.
    angles : numpy.ndarray, optional
        Writes the node angles for each node when included.
    mode : str, default: "w"
        Write mode.
    sig_digits : int, default: 13
        Number of significant digits to use when writing the nodes. Must be
        greater than 0.

    Examples
    --------
    Write random points as nodes for MAPDL.

    >>> import numpy as np
    >>> import mapdl_archive
    >>> points = np.random.random((100, 3))
    >>> point_ids = np.arange(1, 101)
    >>> mapdl_archive.write_nblock("nblock.inp", point_ids, points)

    """
    if sig_digits < 1:
        raise ValueError(f"`sig_digits` must be greater than 0, got {sig_digits}")
    if pos.ndim != 2 or pos.shape[1] != 3:
        raise ValueError(
            f"Invalid position array shape {pos.shape}. Should be shaped `(n, 3)`."
        )
    if angles is not None:
        if angles.ndim != 2 or angles.shape[1] != 3:
            raise ValueError(
                f"Invalid angles array shape {angles.shape}. Should be shaped `(n, 3)`."
            )

    node_id_32: npt.NDArray[np.int32] = node_id.astype(np.int32, copy=False)

    # node array must be sorted
    # note, this is sort check is most suited for pre-sorted arrays
    # see https://stackoverflow.com/questions/3755136/
    if not np.all(node_id_32[:-1] <= node_id_32[1:]):
        sidx = np.argsort(node_id_32)
        node_id_32 = node_id_32[sidx]
        pos = pos[sidx]

    if angles is not None:
        angles = angles.astype(pos.dtype, copy=False)
    else:
        angles = np.empty((0, 0), dtype=pos.dtype)

    if pos.dtype == np.float32 and angles.dtype == np.float32:
        _archive.py_write_nblock_float(
            filename,
            node_id_32,
            node_id_32[-1],
            cast(NPArray_FLOAT32, pos),
            cast(NPArray_FLOAT32, angles),
            mode,
            sig_digits,
        )
    elif pos.dtype == np.float64 and angles.dtype == np.float64:
        _archive.py_write_nblock(
            filename,
            node_id_32,
            node_id_32[-1],
            cast(NPArray_FLOAT64, pos),
            cast(NPArray_FLOAT64, angles),
            mode,
            sig_digits,
        )
    else:
        raise ValueError(
            "Position and angle arrays must both be either float32 or float64"
        )

    return None


def write_cmblock(
    filename: Union[str, io.TextIOBase],
    items: Sequence[int],
    comp_name: str,
    comp_type: str,
    digit_width: int = 10,
    mode: str = "w",
) -> None:
    """Write a component block (CMBLOCK) to a file.

    Parameters
    ----------
    filename : str or file handle
        File to write CMBLOCK component to.
    items : sequence
        Element or node numbers to write.
    comp_name : str
        Name of the component.
    comp_type : str
        Component type to write.  Should be either ``'ELEM'`` or ``'NODE'``.
    digit_width : int, default: 10
        Digit width.
    mode : str, default: "w"
        Write mode.

    Examples
    --------
    Write a node component to disk.

    >>> import write_cmblock
    >>> items = [1, 20, 50, 51, 52, 53]
    >>> mapdl_archive.write_cmblock("cmblock_elem.inp", items, "MY_NODE_COMP", "NODE")

    Write an element component to disk.

    >>> mapdl_archive.write_cmblock("cmblock_elem.inp", items, "MY_NODE_COMP", "ELEM")

    """
    comp_name = comp_name.upper()
    comp_type = comp_type.upper()
    comp_type = "ELEM" if comp_type == "ELEMENT" else comp_type
    if comp_type.upper() not in ["ELEM", "NODE"]:
        raise ValueError("`comp_type` must be either 'ELEM' or 'NODE'")

    opened_file = False
    if isinstance(filename, io.TextIOBase):
        fid = filename
    elif isinstance(filename, str):
        fid = open(filename, mode)  # type: ignore
        opened_file = True
    else:
        raise TypeError(
            "Invalid type {type(filename)} for `filename`. Should be file handle or string."
        )

    if not isinstance(items, np.ndarray):
        items_arr: npt.NDArray[np.int32] = np.array(items, dtype=np.int32)
    else:
        items_arr = items.astype(np.int32, copy=False)

    # All this python writing could be a bottleneck for non-contiguous CMBLOCKs.
    # consider cythonizing this in the future
    cmblock_items: npt.NDArray[np.int32] = _archive.cmblock_items_from_array(items_arr)
    nitems: int = len(cmblock_items)
    print(f"CMBLOCK,{comp_name},{comp_type},{nitems:8d}", file=fid)
    print(f"(8i{digit_width})", file=fid)
    digit_formatter = f"%{digit_width}d"

    # use np savetxt here as it's faster than looping through and
    # writing each line.
    # nearest multiple of 8
    up_to = len(cmblock_items) % 8
    if up_to:  # deal with the zero case
        np.savetxt(fid, cmblock_items[:-up_to].reshape(-1, 8), digit_formatter * 8)

        # write the final line
        chunk = cmblock_items[-up_to:]
        print("".join([digit_formatter] * len(chunk)) % tuple(chunk), file=fid)
    else:
        np.savetxt(fid, cmblock_items.reshape(-1, 8), digit_formatter * 8)

    if opened_file:
        fid.close()


def _write_eblock(
    filename: str,
    elem_id: np.ndarray,
    etype: np.ndarray,
    mtype: np.ndarray,
    rcon: np.ndarray,
    elem_nnodes: np.ndarray,
    cells: np.ndarray,
    offset: np.ndarray,
    celltypes: np.ndarray,
    typenum: np.ndarray,
    nodenum: np.ndarray,
    mode="a",
):
    """Write EBLOCK to disk."""
    # perform type checking here
    elem_id = elem_id.astype(np.int32, copy=False)
    etype = etype.astype(np.int32, copy=False)
    mtype = mtype.astype(np.int32, copy=False)
    rcon = rcon.astype(np.int32, copy=False)
    elem_nnodes = elem_nnodes.astype(np.int32, copy=False)
    cells = cells.astype(np.int32, copy=False)
    offset = offset.astype(np.int32, copy=False)
    celltypes = celltypes.astype(np.uint8, copy=False)
    typenum = typenum.astype(np.int32, copy=False)
    nodenum = nodenum.astype(np.int32, copy=False)

    _archive.py_write_eblock(
        filename,
        elem_id,
        etype,
        mtype,
        rcon,
        elem_nnodes,
        cells,
        offset,
        celltypes,
        typenum,
        nodenum,
        mode=mode,
    )



import numpy as np
import numpy.typing as npt

def py_write_nblock(
    filename: str,
    node_id: npt.NDArray[np.int32],
    max_node_id: int,
    pos: npt.NDArray[np.float64],
    angles: npt.NDArray[np.float64],
    mode: str = "w",
    sig_digits: int = 13,
) -> None: ...
def py_write_nblock_float(
    filename: str,
    node_id: npt.NDArray[np.int32],
    max_node_id: int,
    pos: npt.NDArray[np.float32],
    angles: npt.NDArray[np.float32],
    mode: str = "w",
    sig_digits: int = 13,
) -> None: ...
def py_write_eblock(
    filename: str,
    elem_id: npt.NDArray[np.int32],
    etype: npt.NDArray[np.int32],
    mtype: npt.NDArray[np.int32],
    rcon: npt.NDArray[np.int32],
    elem_nnodes: npt.NDArray[np.int32],
    cells: npt.NDArray[np.int32],
    offset: npt.NDArray[np.int32],
    celltypes: npt.NDArray[np.uint8],
    typenum: npt.NDArray[np.int32],
    nodenum: npt.NDArray[np.int32],
    mode: str = "w",
) -> None: ...
def cmblock_items_from_array(array: npt.NDArray[np.int32]) -> npt.NDArray[np.int32]: ...



"""
Elements available to MAPDL 2020R2.

LINK1
PLANE2
BEAM3
BEAM4
SOLID5
COMBIN7
LINK8
INFIN9
LINK10
LINK11
CONTAC12
PLANE13
COMBIN14
FLUID15
PIPE16
PIPE17
PIPE18
SURF19
PIPE20
MASS21
SURF22
BEAM23
BEAM24
PLANE25
MATRIX27
SHELL28
FLUID29
FLUID30
LINK31
LINK32
LINK33
LINK34
PLANE35
SOURC36
COMBIN37
FLUID38
COMBIN39
COMBIN40
SHELL41
PLANE42
SHELL43
BEAM44
SOLID45
SOLID46
INFIN47
MATRIX50
SHELL51
CONTAC52
PLANE53
BEAM54
PLANE55
SHELL57
PIPE59
PIPE60
SHELL61
SOLID62
SHELL63
SOLID64
SOLID65
FLUID66
PLANE67
LINK68
SOLID69
SOLID70
MASS71
PLANE75
PLANE77
PLANE78
FLUID79
FLUID80
FLUID81
PLANE82
PLANE83
SOLID87
VISCO88
VISCO89
SOLID90
SHELL91
SOLID92
SHELL93
CIRCU94
SOLID95
SOLID96
SOLID97
SOLID98
SHELL99
USER100
USER101
USER102
USER103
USER104
USER105
VISCO106
VISCO107
VISCO108
TRANS109
INFIN110
INFIN111
ROT112
INTER115
FLUID116
EDGE117
HF118
HF119
HF120
PLANE121
SOLID122
SOLID123
CIRCU124
CIRCU125
TRANS126
FLUID129
FLUID130
SHELL131
SHELL132
TRANS135
FLUID136
FLUID137
FLUID138
FLUID139
ROM140
SHELL143
ROM144
SURF151
SURF152
SURF153
SURF154
SURF155
SURF156
SHELL157
SURF159
TARGE169
TARGE170
CONTA171
CONTA172
CONTA173
CONTA174
CONTA175
CONTA176
CONTA177
CONTA178
PRETS179
LINK180
SHELL181
PLANE182
PLANE183
RBAR184
SOLID185
SOLID186
SOLID187
BEAM188
BEAM189
SOLID190
SOLID191
INTER192
INTER193
INTER194
INTER195
LAYER196
LAYER197
FOLLW201
INTER202
INTER203
INTER204
INTER205
INTER206
SHELL208
SHELL209
CPT212
CPT213
COMBI214
CPT215
CPT216
CPT217
FLUID218
FLUID219
FLUID220
FLUID221
PLANE222
SOLID223
SOLID225
SOLID226
SOLID227
PLANE230
SOLID231
SOLID232
PLANE233
SOLID236
SOLID237
PLANE238
SOLID239
SOLID240
HSFLD241
HSFLD242
COMBI250
SURF251
SURF252
INFIN257
INFIN258
INFIN259
INFIN260
GLINK262
REINF263
REINF264
REINF265
SOLID272
SOLID273
SOLID278
SOLID279
CABLE280
SHELL281
SHELL282
SHELL283
SOLID285
PIPE288
PIPE289
ELBOW290
SOLID291
PLANE292
PLANE293
USER300
"""

from typing import List

import numpy as np
import numpy.typing as npt

# element type to VTK conversion function call map
# 0: skip
# 1: Point
# 2: Line (linear or quadratic)
# 3: Shell
# 4: 3D Solid (Hexahedral, wedge, pyramid, tetrahedral)
# 5: Tetrahedral
# 6: Line (always linear)

_etype_map: List[int] = [
    0,
    2,  # LINK1
    3,  # PLANE2
    2,  # BEAM3
    2,  # BEAM4
    4,  # SOLID5
    0,  # UNUSED6
    1,  # COMBIN7
    2,  # LINK8
    0,  # INFIN9
    2,  # LINK10
    2,  # LINK11
    2,  # CONTAC12
    3,  # PLANE13
    2,  # COMBIN14
    0,  # FLUID15
    2,  # PIPE16
    2,  # PIPE17
    2,  # PIPE18
    0,  # SURF19
    2,  # PIPE20
    1,  # MASS21
    3,  # SURF22
    2,  # BEAM23
    2,  # BEAM24
    3,  # PLANE25
    0,  # UNUSED26
    0,  # MATRIX27
    3,  # SHELL28
    3,  # FLUID29
    4,  # FLUID30
    2,  # LINK31
    2,  # LINK32
    2,  # LINK33
    2,  # LINK34
    3,  # PLANE35
    0,  # SOURC36
    2,  # COMBIN37
    2,  # FLUID38
    2,  # COMBIN39
    2,  # COMBIN40
    3,  # SHELL41
    3,  # PLANE42
    3,  # SHELL43
    2,  # BEAM44
    4,  # SOLID45
    4,  # SOLID46
    0,  # INFIN47
    0,  # UNUSED48
    0,  # UNUSED49
    0,  # MATRIX50
    3,  # SHELL51
    0,  # CONTAC52
    3,  # PLANE53
    3,  # BEAM54
    3,  # PLANE55
    0,  # UNUSED56
    3,  # SHELL57
    0,  # UNUSED58
    2,  # PIPE59
    2,  # PIPE60
    2,  # SHELL61
    4,  # SOLID62
    3,  # SHELL63
    4,  # SOLID64
    4,  # SOLID65
    2,  # FLUID66
    3,  # PLANE67
    2,  # LINK68
    4,  # SOLID69
    4,  # SOLID70
    1,  # MASS71
    0,  # UNUSED72
    0,  # UNUSED73
    0,  # UNUSED74
    3,  # PLANE75
    0,  # UNUSED76
    3,  # PLANE77
    3,  # PLANE78
    3,  # FLUID79
    4,  # FLUID80
    3,  # FLUID81
    3,  # PLANE82
    3,  # PLANE83
    0,  # UNUSED84
    0,  # UNUSED85
    0,  # UNUSED86
    5,  # SOLID87
    3,  # VISCO88
    4,  # VISCO89
    4,  # SOLID90
    3,  # SHELL91
    5,  # SOLID92
    3,  # SHELL93
    0,  # CIRCU94
    4,  # SOLID95
    4,  # SOLID96
    4,  # SOLID97
    5,  # SOLID98
    3,  # SHELL99
    4,  # USER100
    4,  # USER101
    4,  # USER102
    4,  # USER103
    4,  # USER104
    4,  # USER105
    3,  # VISCO106
    4,  # VISCO107
    4,  # VISCO108
    0,  # TRANS109
    0,  # INFIN110
    0,  # INFIN111
    0,  # ROT112
    0,  # UNUSED113
    0,  # UNUSED114
    3,  # INTER115
    2,  # FLUID116
    4,  # EDGE117
    3,  # HF118
    5,  # HF119
    4,  # HF120
    3,  # PLANE121
    4,  # SOLID122
    5,  # SOLID123
    0,  # CIRCU124
    0,  # CIRCU125
    2,  # TRANS126
    0,  # UNUSED127
    0,  # UNUSED128
    2,  # FLUID129
    3,  # FLUID130
    3,  # SHELL131
    3,  # SHELL132
    0,  # UNUSED133
    0,  # UNUSED134
    0,  # TRANS135
    3,  # FLUID136
    0,  # FLUID137
    0,  # FLUID138
    0,  # FLUID139
    5,  # ROM140
    0,  # UNUSED141
    0,  # UNUSED142
    3,  # SHELL143
    0,  # ROM144
    0,  # UNUSED145
    0,  # UNUSED146
    0,  # UNUSED147
    0,  # UNUSED148
    0,  # UNUSED149
    0,  # UNUSED150
    2,  # SURF151
    3,  # SURF152
    2,  # SURF153
    3,  # SURF154
    3,  # SURF155
    2,  # SURF156
    3,  # SHELL157
    0,  # UNUSED158
    0,  # SURF159
    0,  # UNUSED160
    2,  # BEAM161
    0,  # UNUSED162
    3,  # SHELL163
    4,  # SOLID164
    0,  # UNUSED165
    0,  # UNUSED166
    0,  # UNUSED167
    5,  # SOLID168
    0,  # TARGE169
    3,  # TARGE170
    2,  # CONTA171
    2,  # CONTA172
    3,  # CONTA173
    3,  # CONTA174
    1,  # CONTA175
    2,  # CONTA176
    2,  # CONTA177
    2,  # CONTA178
    0,  # PRETS179
    2,  # LINK180
    3,  # SHELL181
    3,  # PLANE182
    3,  # PLANE183
    0,  # RBAR184
    4,  # SOLID185
    4,  # SOLID186
    5,  # SOLID187
    6,  # BEAM188
    2,  # BEAM189
    4,  # SOLSH190
    0,  # UNUSED191
    4,  # INTER192
    0,  # INTER193
    0,  # INTER194
    0,  # INTER195
    0,  # LAYER196
    0,  # LAYER197
    0,  # UNUSED198
    0,  # UNUSED199
    0,  # UNUSED200
    0,  # FOLLW201
    0,  # INTER202
    0,  # INTER203
    0,  # INTER204
    0,  # INTER205
    0,  # INTER206
    0,  # UNUSED207
    2,  # SHELL208
    2,  # SHELL209
    0,  # UNUSED210
    0,  # UNUSED211
    0,  # CPT212
    0,  # CPT213
    0,  # COMBI214
    0,  # CPT215
    0,  # UNUSED216
    5,  # CPT217
    3,  # FLUID218
    3,  # FLUID219
    4,  # FLUID220
    5,  # FLUID221
    3,  # PLANE222
    3,  # PLANE223
    0,  # UNUSED224
    0,  # SOLID225
    4,  # SOLID226
    5,  # SOLID227
    0,  # UNUSED228
    0,  # UNUSED229
    3,  # PLANE230
    4,  # SOLID231
    5,  # SOLID232
    3,  # PLANE233
    0,  # UNUSED234
    0,  # UNUSED235
    4,  # SOLID236
    5,  # SOLID237
    3,  # PLANE238
    4,  # SOLID239
    5,  # SOLID240
    0,  # HSFLD241
    0,  # HSFLD242
    0,  # UNUSED243
    0,  # UNUSED244
    0,  # UNUSED245
    0,  # UNUSED246
    0,  # UNUSED247
    0,  # UNUSED248
    0,  # UNUSED249
    2,  # COMBI250
    2,  # SURF251
    3,  # SURF252
    0,  # UNUSED253
    0,  # UNUSED254
    0,  # UNUSED255
    0,  # UNUSED256
    0,  # INFIN257
    0,  # INFIN258
    0,  # INFIN259
    0,  # UNUSED260
    4,  # UNUSED261
    0,  # GLINK262
    0,  # REINF263
    0,  # REINF264
    0,  # REINF265
    0,  # UNUSED266
    0,  # UNUSED267
    0,  # UNUSED268
    0,  # UNUSED269
    0,  # UNUSED270
    0,  # UNUSED271
    0,  # SOLID272
    0,  # SOLID273
    0,  # UNUSED274
    0,  # UNUSED275
    0,  # UNUSED276
    0,  # UNUSED277
    4,  # SOLID278
    4,  # SOLID279
    2,  # CABLE280
    3,  # SHELL281
    3,  # SHELL282
    3,  # SHELL283
    0,  # UNUSED284
    5,  # SOLID285
    0,  # UNUSED286
    0,  # UNUSED287
    2,  # PIPE288
    2,  # PIPE289
    2,  # ELBOW290
    5,  # SOLID291
    3,  # PLANE292
    3,  # PLANE293
    0,  # UNUSED294
    0,  # UNUSED295
    0,  # UNUSED296
    0,  # UNUSED297
    0,  # UNUSED298
    0,  # UNUSED299
    0,  # USER300
]

ETYPE_MAP: npt.NDArray[np.int32] = np.array(_etype_map, dtype=np.int32)

# this is a mask of solid elements
SOLID_ELEMENTS: npt.NDArray[np.bool_] = np.array(
    [
        False,  # **dummy value**
        False,  # LINK1
        False,  # PLANE2
        False,  # BEAM3
        False,  # BEAM4
        True,  # SOLID5
        False,  # UNUSED6
        False,  # COMBIN7
        False,  # LINK8
        False,  # INFIN9
        False,  # LINK10
        False,  # LINK11
        False,  # CONTAC12
        False,  # PLANE13
        False,  # COMBIN14
        False,  # FLUID15
        False,  # PIPE16
        False,  # PIPE17
        False,  # PIPE18
        False,  # SURF19
        False,  # PIPE20
        False,  # MASS21
        False,  # SURF22
        False,  # BEAM23
        False,  # BEAM24
        False,  # PLANE25
        False,  # UNUSED26
        False,  # MATRIX27
        False,  # SHELL28
        False,  # FLUID29
        False,  # FLUID30
        False,  # LINK31
        False,  # LINK32
        False,  # LINK33
        False,  # LINK34
        False,  # PLANE35
        False,  # SOURC36
        False,  # COMBIN37
        False,  # FLUID38
        False,  # COMBIN39
        False,  # COMBIN40
        False,  # SHELL41
        False,  # PLANE42
        False,  # SHELL43
        False,  # BEAM44
        True,  # SOLID45
        True,  # SOLID46
        False,  # INFIN47
        False,  # UNUSED48
        False,  # UNUSED49
        False,  # MATRIX50
        False,  # SHELL51
        False,  # CONTAC52
        False,  # PLANE53
        False,  # BEAM54
        False,  # PLANE55
        False,  # UNUSED56
        False,  # SHELL57
        False,  # UNUSED58
        False,  # PIPE59
        False,  # PIPE60
        False,  # SHELL61
        True,  # SOLID62
        False,  # SHELL63
        True,  # SOLID64
        True,  # SOLID65
        False,  # FLUID66
        False,  # PLANE67
        False,  # LINK68
        True,  # SOLID69
        True,  # SOLID70
        False,  # MASS71
        False,  # UNUSED72
        False,  # UNUSED73
        False,  # UNUSED74
        False,  # PLANE75
        False,  # UNUSED76
        False,  # PLANE77
        False,  # PLANE78
        False,  # FLUID79
        False,  # FLUID80
        False,  # FLUID81
        False,  # PLANE82
        False,  # PLANE83
        False,  # UNUSED84
        False,  # UNUSED85
        False,  # UNUSED86
        True,  # SOLID87
        False,  # VISCO88
        False,  # VISCO89
        True,  # SOLID90
        False,  # SHELL91
        True,  # SOLID92
        False,  # SHELL93
        False,  # CIRCU94
        True,  # SOLID95
        True,  # SOLID96
        True,  # SOLID97
        True,  # SOLID98
        False,  # SHELL99
        False,  # USER100
        False,  # USER101
        False,  # USER102
        False,  # USER103
        False,  # USER104
        False,  # USER105
        False,  # VISCO106
        False,  # VISCO107
        False,  # VISCO108
        False,  # TRANS109
        False,  # INFIN110
        False,  # INFIN111
        False,  # ROT112
        False,  # UNUSED113
        False,  # UNUSED114
        False,  # INTER115
        False,  # FLUID116
        False,  # EDGE117
        False,  # HF118
        False,  # HF119
        False,  # HF120
        False,  # PLANE121
        True,  # SOLID122
        True,  # SOLID123
        False,  # CIRCU124
        False,  # CIRCU125
        False,  # TRANS126
        False,  # UNUSED127
        False,  # UNUSED128
        False,  # FLUID129
        False,  # FLUID130
        False,  # SHELL131
        False,  # SHELL132
        False,  # UNUSED133
        False,  # UNUSED134
        False,  # TRANS135
        False,  # FLUID136
        False,  # FLUID137
        False,  # FLUID138
        False,  # FLUID139
        False,  # ROM140
        False,  # UNUSED141
        False,  # UNUSED142
        False,  # SHELL143
        False,  # ROM144
        False,  # UNUSED145
        False,  # UNUSED146
        False,  # UNUSED147
        False,  # UNUSED148
        False,  # UNUSED149
        False,  # UNUSED150
        False,  # SURF151
        False,  # SURF152
        False,  # SURF153
        False,  # SURF154
        False,  # SURF155
        False,  # SURF156
        False,  # SHELL157
        False,  # UNUSED158
        False,  # SURF159
        False,  # UNUSED160
        False,  # BEAM161
        False,  # UNUSED162
        False,  # SHELL163
        True,  # SOLID164
        False,  # UNUSED165
        False,  # UNUSED166
        False,  # UNUSED167
        True,  # SOLID168
        False,  # TARGE169
        False,  # TARGE170
        False,  # CONTA171
        False,  # CONTA172
        False,  # CONTA173
        False,  # CONTA174
        False,  # CONTA175
        False,  # CONTA176
        False,  # CONTA177
        False,  # CONTA178
        False,  # PRETS179
        False,  # LINK180
        False,  # SHELL181
        False,  # PLANE182
        False,  # PLANE183
        False,  # RBAR184
        True,  # SOLID185
        True,  # SOLID186
        True,  # SOLID187
        False,  # BEAM188
        False,  # BEAM189
        False,  # SOLSH190
        False,  # UNUSED191
        False,  # INTER192
        False,  # INTER193
        False,  # INTER194
        False,  # INTER195
        False,  # LAYER196
        False,  # LAYER197
        False,  # UNUSED198
        False,  # UNUSED199
        False,  # UNUSED200
        False,  # FOLLW201
        False,  # INTER202
        False,  # INTER203
        False,  # INTER204
        False,  # INTER205
        False,  # INTER206
        False,  # UNUSED207
        False,  # SHELL208
        False,  # SHELL209
        False,  # UNUSED210
        False,  # UNUSED211
        False,  # CPT212
        False,  # CPT213
        False,  # COMBI214
        False,  # CPT215
        False,  # UNUSED216
        False,  # CPT217
        False,  # FLUID218
        False,  # FLUID219
        False,  # FLUID220
        False,  # FLUID221
        False,  # PLANE222
        False,  # PLANE223
        False,  # UNUSED224
        True,  # SOLID225
        True,  # SOLID226
        True,  # SOLID227
        False,  # UNUSED228
        False,  # UNUSED229
        False,  # PLANE230
        True,  # SOLID231
        True,  # SOLID232
        False,  # PLANE233
        False,  # UNUSED234
        False,  # UNUSED235
        True,  # SOLID236
        True,  # SOLID237
        False,  # PLANE238
        True,  # SOLID239
        True,  # SOLID240
        False,  # HSFLD241
        False,  # HSFLD242
        False,  # UNUSED243
        False,  # UNUSED244
        False,  # UNUSED245
        False,  # UNUSED246
        False,  # UNUSED247
        False,  # UNUSED248
        False,  # UNUSED249
        False,  # COMBI250
        False,  # SURF251
        False,  # SURF252
        False,  # UNUSED253
        False,  # UNUSED254
        False,  # UNUSED255
        False,  # UNUSED256
        False,  # INFIN257
        False,  # INFIN258
        False,  # INFIN259
        False,  # UNUSED260
        False,  # UNUSED261
        False,  # GLINK262
        False,  # REINF263
        False,  # REINF264
        False,  # REINF265
        False,  # UNUSED266
        False,  # UNUSED267
        False,  # UNUSED268
        False,  # UNUSED269
        False,  # UNUSED270
        False,  # UNUSED271
        True,  # SOLID272
        True,  # SOLID273
        False,  # UNUSED274
        False,  # UNUSED275
        False,  # UNUSED276
        False,  # UNUSED277
        True,  # SOLID278
        True,  # SOLID279
        False,  # CABLE280
        False,  # SHELL281
        False,  # SHELL282
        False,  # SHELL283
        False,  # UNUSED284
        True,  # SOLID285
        False,  # UNUSED286
        False,  # UNUSED287
        False,  # PIPE288
        False,  # PIPE289
        False,  # ELBOW290
        True,  # SOLID291
        False,  # PLANE292
        False,  # PLANE293
        False,  # UNUSED294
        False,  # UNUSED295
        False,  # UNUSED296
        False,  # UNUSED297
        False,  # UNUSED298
        False,  # UNUSED299
        False,  # USER300
    ]
)



"""Mapdl-archive example files."""

import os

# get location of this folder and the example files
dir_path = os.path.dirname(os.path.realpath(__file__))
hexarchivefile = os.path.join(dir_path, "HexBeam.cdb")
tetarchivefile = os.path.join(dir_path, "TetBeam.cdb")
sector_archive_file = os.path.join(dir_path, "sector.cdb")
academic_rotor_archive_file = os.path.join(dir_path, "academic_rotor.rst")



from typing import Dict, List, Tuple, TypedDict, Union

import numpy as np
import numpy.typing as npt

class ReadReturnDict(TypedDict):
    rnum: npt.NDArray[np.int64]
    rdat: List[List[float]]
    ekey: npt.NDArray[np.int32]
    nnum: npt.NDArray[np.int32]
    nodes: npt.NDArray[np.float64]
    elem: npt.NDArray[np.int32]
    elem_off: npt.NDArray[np.int32]
    node_comps: Dict[str, npt.NDArray[np.int32]]
    elem_comps: Dict[str, npt.NDArray[np.int32]]
    keyopt: Dict[int, List[List[int]]]
    parameters: Dict[str, np.ndarray]
    nblock_start: int
    nblock_end: int

def read(
    filename: str,
    read_parameters: bool = False,
    debug: bool = False,
    read_eblock: bool = True,
) -> ReadReturnDict: ...
def node_block_format(
    string: Union[bytes, str],
) -> Tuple[npt.NDArray[np.int32], int, int, int]: ...
def ans_vtk_convert(
    elem: npt.NDArray[np.int32],
    elem_off: npt.NDArray[np.int32],
    type_ref: npt.NDArray[np.int32],
    nnum: npt.NDArray[np.int32],
    build_offset: int,
) -> Tuple[npt.NDArray[np.int64], npt.NDArray[np.uint8], npt.NDArray[np.int64]]: ...
def read_from_nwrite(
    filename: str, nnodes: int
) -> Tuple[npt.NDArray[np.int32], npt.NDArray[np.float64]]: ...
def write_array(filename: str, arr: npt.NDArray[np.float64]) -> None: ...



# cython: language_level=2
# cython: infer_types=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: cdivision=True
# cython: nonecheck=False
# cython: embedsignature=True

""" Cython implementation of a CDB reader """
from libc.stdint cimport int64_t
from libc.stdio cimport (
    EOF,
    FILE,
    SEEK_CUR,
    SEEK_END,
    SEEK_SET,
    fclose,
    fgetc,
    fgets,
    fopen,
    fread,
    fscanf,
    fseek,
    ftell,
    printf,
    sscanf,
    ungetc,
)
from libc.stdlib cimport atof, atoi, free, malloc
from libc.string cimport memcmp, strcmp, strncpy

ctypedef unsigned char uint8_t

import ctypes

import numpy as np

cimport numpy as np


def safe_int(value):
    """Safely convert a value to int.

    Return int if can be converted, None if otherwise
    """
    try:
        return int(value)
    except:
        return


cdef extern from "reader.h":
    int read_nblock_from_nwrite(char*, int*, double*, int)
    int write_array_ascii(const char*, const double*, int);
    int read_eblock_cfile(FILE *cfile, int *elem_off, int *elem, int nelem)
    int read_nblock_cfile(FILE *cfile, int *nnum, double *nodes, int nnodes, int* d_size, int f_size)

cdef extern from 'vtk_support.h':
    int ans_to_vtk(const int, const int*, const int*, const int*, const int,
    const int*, int64_t*, int64_t*, uint8_t*, const int)


def read(filename, read_parameters=False, debug=False, read_eblock=True):
    """Read blocked ansys archive file."""
    badstr = 'Badly formatted cdb file'
    filename_byte_string = filename.encode("UTF-8")
    cdef char* fname = filename_byte_string
    parameters = {}
    # Check file exists
    cdef FILE* cfile
    cfile = fopen(fname, 'r')

    if cfile == NULL:
        raise Exception("No such file or directory: '%s'" % filename)

    # Define variables
    cdef size_t l = 0
    cdef ssize_t read
    cdef int[5] blocksz
    cdef int i, j, linelen, isz, tempint
    cdef float tempflt

    # Size temp char array
    cdef char line[500]
    cdef char tempstr[100]

    # Get element types
    elem_type = []
    rnum = []
    rdat = []

    # NBLOCK
    cdef int nnodes = 0
    cdef int [::1] nnum = np.empty(0, ctypes.c_int)
    cdef double [:, ::1] nodes = np.empty((0, 0))
    cdef int d_size[3]
    cdef int f_size, nfields, nblock_start, nblock_end, _start

    # EBLOCK
    cdef int nelem = 0
    cdef int [::1] elem = np.empty(0, dtype=ctypes.c_int)
    cdef int [::1] elem_off = np.empty(0, dtype=ctypes.c_int)
    cdef int elem_sz

    # CMBLOCK
    cdef int ncomp
    cdef int [::1] component
    cdef int nblock

    node_comps = {}
    elem_comps = {}

    nodes_read = False
    eblock_read = False

    # parameters
    cdef char *end_marker = b"END PREAD"
    cdef list parm_lines = []
    cdef long position

    # keyopt
    keyopt = {}

    cdef int first_char, next_char;

    while 1:
        first_char = fgetc(cfile)
        if first_char == EOF:
            break
        elif first_char == '\r' or first_char == '\n':
            ungetc(first_char, cfile)
            fgets(line, sizeof(line), cfile)
        elif first_char == 'E' or first_char == 'e':
            ungetc(first_char, cfile)
            fgets(line, sizeof(line), cfile)

            # Record element types
            if b'ET,' == line[:3] or b'et,' == line[:3]:
                if debug:
                    print('reading ET')

                # element number and element type
                et_val = line.decode().split(',')
                try:
                    int(et_val[1])
                    elem_type.append([int(et_val[1]), int(et_val[2])])
                except:
                    if debug:
                        print('Invalid "ET" command %s' % line.decode())
                    continue

            # read in new ETBLOCK (replacement for ET)
            elif b'ETBLOCK' in line or b'etblock' in line:
                # read the number of items in the block
                set_dat = [safe_int(value) for value in line.split(b',')[1:]]
                n_items = set_dat[0]

                # Skip Format1 (2i9,19a9)
                fgets(line, sizeof(line), cfile)

                for item in range(n_items):
                    fgets(line, sizeof(line), cfile)

                    # Only read in the first two items (ELEM index and TYPE)
                    # NOTE: Remaining contents of the block are unknown
                    et_val = [safe_int(value) for value in line.split()[:2]]
                    elem_type.append(et_val)

            elif b'EBLOCK,' == line[:7] or b'eblock,' == line[:7] and read_eblock:
                if eblock_read:
                    # Sometimes, DAT files contain two EBLOCKs.  Read
                    # only the first block.
                    if debug:
                        print('EBLOCK already read, skipping...')
                    continue
                if debug:
                    print('reading EBLOCK...')

                # only read entries with SOLID
                if b'SOLID' in line or b'solid' in line:

                    # Get size of EBLOCK from the last item in the line
                    # Example: "EBLOCK,19,SOLID,,3588"
                    nelem = int(line[line.rfind(b',') + 1:])
                    if nelem == 0:
                        raise RuntimeError('Unable to read element block')

                    # Populate element field data and connectivity
                    elem = np.empty(nelem*30, dtype=ctypes.c_int)
                    elem_off = np.empty(nelem + 1, dtype=ctypes.c_int)
                    elem_sz = read_eblock_cfile(cfile, &elem_off[0], &elem[0], nelem)

                    eblock_read = True
                    if debug:
                        print('finished')

        elif first_char == 'K' or first_char == 'k':
            ungetc(first_char, cfile);  # Put the character back into the stream
            fgets(line, sizeof(line), cfile)

            if debug:
                print('Hit "K"')

            if b'KEYOP' in line or b'keyop' in line:
                if debug:
                    print('reading KEYOP')

                try:
                    entry = []
                    for item in line.split(b',')[1:]:
                        entry.append(int(item))
                except:
                    continue

                key_num = int(entry[0])
                if key_num in keyopt:
                    keyopt[key_num].append(entry[1:])
                else:
                    keyopt[key_num] = [entry[1:]]

        elif first_char == 'R' or first_char == 'r':
            ungetc(first_char, cfile);  # Put the character back into the stream
            fgets(line, sizeof(line), cfile)

            if debug:
                print('Hit "R"')

            if b'RLBLOCK' in line or b'rlblock' in line:
                if debug:
                    print('reading RLBLOCK')

                # The RLBLOCK command defines a real constant
                # set. The real constant sets follow each set,
                # starting with

                # Format1 and followed by one or more Format2's, as
                # needed. The command format is:
                # RLBLOCK,NUMSETS,MAXSET,MAXITEMS,NPERLINE
                # Format1
                # Format2
                #
                # where:
                # Format1 - Data descriptor defining the format of the
                # first line. For the RLBLOCK command, this is always
                # (2i8,6g16.9).  The first i8 is the set number, the
                # second i8 is the number of values in this set,
                # followed by up to 6 real
                #
                # Format2 - Data descriptors defining the format of
                # the subsequent lines (as needed); this is always
                # (7g16.9).
                #
                # - NUMSETS : The number of real constant sets defined
                # - MAXSET : Maximum real constant set number
                # - MAXITEMS : Maximum number of reals in any one set
                # - NPERLINE : Number of reals defined on a line

                # Get number of sets from the RLBLOCK line
                set_dat = [safe_int(value) for value in line.split(b',')[1:]]
                nset, maxset, maxitems, nperline = set_dat

                # Skip Format1 and Format2 (always 2i8,6g16.9 and 7g16.9)
                if fgets(line, sizeof(line), cfile) == NULL: raise RuntimeError(badstr)
                if fgets(line, sizeof(line), cfile) == NULL: raise RuntimeError(badstr)

                # Read data
                for _ in range(nset):
                    rcon = [] # real constants

                    if fgets(line, sizeof(line), cfile) == NULL: raise RuntimeError(badstr)

                    # Get real constant number
                    rnum.append(int(line[:8]))

                    # Number of constants
                    ncon = int(line[8:16])

                    # Get constant data
                    if ncon > 6: # if multiple lines
                        for i in range(6):
                            rcon.append(float(line[16 + 16*i:32 + 16*i]))
                            ncon -= 1

                        # advance line
                        if fgets(line, sizeof(line), cfile) == NULL:
                            raise RuntimeError(badstr)

                        # read next line
                        while True:
                            if ncon > 7:
                                for i in range(7):
                                    rcon.append(float(line[16*i:16*(i + 1)]))
                                    ncon -= 1
                                # advance
                                if fgets(line, sizeof(line), cfile) == NULL:
                                    raise RuntimeError(badstr)

                            else:
                                for i in range(ncon):
                                    try:
                                        rcon.append(float(line[16*i:16 + 16*i]))
                                    # account for empty 0 values
                                    except:
                                        rcon.append(0.0)

                                break

                    # If only one in constant data
                    else:
                        for i in range(ncon):
                            rcon.append(float(line[16 + 16*i:32 + 16*i]))

                    rdat.append(rcon)

        elif first_char == 'N' or first_char == 'n':
            ungetc(first_char, cfile);  # Put the character back into the stream
            _start = ftell(cfile)  # used for NBLOCK
            fgets(line, sizeof(line), cfile)

            if debug:
                print('Hit "N"')

            # if line contains the start of the node block
            if line[:6] == b'NBLOCK' or line[:6] == b'nblock':
                if nodes_read:
                    if debug:
                        print('Skipping additional NBLOCK')
                    continue
                if debug:
                    print('reading NBLOCK due to ', line.decode())

                # Before reading NBLOCK, save where the nblock started
                nblock_start = _start

                # Get size of NBLOCK
                nnodes = int(line[line.rfind(b',') + 1:])
                nnum = np.empty(nnodes, dtype=ctypes.c_int)
                nodes = np.empty((nnodes, 6))

                # Get format of nblock
                fgets(line, sizeof(line), cfile)
                d_size_py, f_size, _, _ = node_block_format(line)
                d_size[0] = d_size_py[0]
                d_size[1] = d_size_py[1]
                d_size[2] = d_size_py[2]

                nnodes_read = read_nblock_cfile(cfile, &nnum[0], &nodes[0, 0], nnodes, d_size, f_size)
                nodes_read = True

                # read final line
                fgets(line, sizeof(line), cfile)
                nblock_end = ftell(cfile)

                if nnodes_read != nnodes:
                    nnodes = nnodes_read
                    nodes = nodes[:nnodes]
                    nnum = nnum[:nnodes]

                if debug:
                    print('Read', nnodes_read, 'nodes')

        elif first_char == 'C' or first_char == 'c':
            ungetc(first_char, cfile);  # Put the character back into the stream
            fgets(line, sizeof(line), cfile)

            if debug:
                print('Hit "C"')

            if line[:8] == b'CMBLOCK,' or line[:8] == b'cmblock,':  # component block
                if debug:
                    print('reading CMBLOCK')

                try:
                    line.decode()
                except:
                    print('Poor formatting of CMBLOCK: %s' % line)
                    continue

                split_line = line.split(b',')
                if len(split_line) < 3:
                    print('Poor formatting of CMBLOCK: %s' % line)
                    continue

                # Get number of items
                ncomp = int(split_line[3].split(b'!')[0].strip())
                component = np.empty(ncomp, ctypes.c_int)

                # Get integer size
                fgets(line, sizeof(line), cfile)
                isz = int(line[line.find(b'i') + 1:line.find(b')')])
                tempstr[isz] = '\0'

                # Number of integers per line
                nblock = int(line[line.find(b'(') + 1:line.find(b'i')])

                # Extract nodes
                for i in range(ncomp):

                    # Read new line if at the end of the line
                    if i%nblock == 0:
                        fgets(line, sizeof(line), cfile)

                    strncpy(tempstr, line + isz*(i%nblock), isz)
                    component[i] = atoi(tempstr)

                # Convert component to array and store
                comname = split_line[1].decode().strip()
                line_comp_type = split_line[2]
                if b'NODE' in line_comp_type:
                    node_comps[comname] = component_interperter(component)
                elif b'ELEM' in line_comp_type:
                    elem_comps[comname] = component_interperter(component)

        elif read_parameters and first_char == '*':  # maybe *DIM
            ungetc(first_char, cfile);  # Put the character back into the stream
            fgets(line, sizeof(line), cfile)

            if debug:
                print('Hit "*"')

            if b'DIM' in line:
                items = line.decode().split(',')
                if len(items) < 3:
                    continue

                name = items[1]
                if items[2].lower() == 'string':
                    fgets(line, sizeof(line), cfile)
                    string_items = line.decode().split('=')
                    if len(string_items) > 1:
                        parameters[name] = string_items[1].replace("'", '').strip()
                    else:
                        parameters[name] = line.decode()
                elif items[2].lower() == 'array':
                    fgets(line, sizeof(line), cfile)
                    if b'PREAD' in line:
                        if debug:
                            print('reading PREAD')

                        _, name, arr_size = line.decode().split(',')
                        name = name.strip()
                        position = ftell(cfile)

                        while True:
                            if fgets(line, sizeof(line), cfile) is NULL:
                                # EOF or error: Reset the file position
                                if fseek(cfile, position, SEEK_SET) != 0:
                                    # Error in fseek
                                    raise IOError("Error setting file position")
                                # EOF or error
                                break

                            # Check for the end marker and end if reached it
                            if memcmp(line, end_marker, sizeof(end_marker) - 1) == 0:
                                break

                            parm_lines.append(line)

                        parameters[name] = np.genfromtxt((b''.join(parm_lines)).split())

        else:
            # no match, simply read remainder of line

            fgets(line, sizeof(line), cfile)

    # # if the node block was not read for some reason
    # if not nodes_read:
    #     if debug:
    #         print('Did not read nodes block. Rereading from start.')

    if debug:
        print('Returning arrays')

    return {
        'rnum': np.asarray(rnum),
        'rdat': rdat,
        'ekey': np.asarray(elem_type, ctypes.c_int),
        'nnum': np.asarray(nnum),
        'nodes': np.asarray(nodes),
        'elem': np.array(elem[:elem_sz]),
        'elem_off': np.array(elem_off),
        'node_comps': node_comps,
        'elem_comps': elem_comps,
        'keyopt': keyopt,
        'parameters': parameters,
        'nblock_start': nblock_start,
        'nblock_end': nblock_end,
    }


def node_block_format(string):
    """Get the node block format.

    Example formats:
    (3i9,6e21.13e3)
    3 ints, all 9 digits wide followed by 6 floats

    (1i7,2i9,6e21.13)
    1 int 7 digits wide, 2 ints, 9 digits wide, 6 floats
    """
    string = string.decode().replace('(', '').replace(')', '')
    fields = string.split(',')

    # double and float size
    d_size = np.zeros(3, np.int32)
    nexp = 2  # default when missing
    nfields = 6
    f_size = 21
    c = 0
    for field in fields:
        if 'i' in field:
            items = field.split('i')
            for n in range(int(items[0])):
                d_size[c] = int(items[1])
                c += 1
        elif 'e' in field:
            f_size = int(field.split('e')[1].split('.')[0])

            # get number of possible integers in the float scientific notation
            if 'e' in field.split('.')[1]:
                nexp = int(field.split('.')[1].split('e')[1])

            nfields = int(field.split('e')[0])

    return d_size, f_size, nfields, nexp


def component_interperter(component):
    """If a node is negative, it is describing a list from the
    previous node.  This is ANSYS's way of saving file size when
    writing components.

    This function has not been optimized.

    """
    f_new = []
    for i in range(len(component)):
        if component[i] > 0: # Append if positive
            f_new.append(component[i])
        else: # otherwise, append list
            f_new.append(range(abs(component[i - 1]) + 1, abs(component[i]) + 1))

    return np.hstack(f_new).astype(ctypes.c_int)


def ans_vtk_convert(const int [::1] elem, const int [::1] elem_off,
                    const int [::1] type_ref,
                    const int [::1] nnum, int build_offset):
    """Convert ansys style connectivity to VTK connectivity"""
    cdef int nelem = elem_off.size - 1
    cdef int64_t [::1] offset = np.empty(nelem, ctypes.c_int64)
    cdef uint8_t [::1] celltypes = np.empty(nelem, dtype='uint8')

    # Allocate connectivity
    # max cell size is 20 (VTK_HEXAHEDRAL) and cell header is 1
    cdef int64_t [::1] cells = np.empty(nelem*21, ctypes.c_int64)
    cdef int loc = ans_to_vtk(nelem, &elem[0], &elem_off[0],
                              &type_ref[0], nnum.size, &nnum[0],
                              &offset[0], &cells[0], &celltypes[0],
                              build_offset)

    return np.asarray(offset), np.asarray(celltypes), np.asarray(cells[:loc])


def read_from_nwrite(filename, int nnodes):
    """Read the node coordinates from the output from the MAPDL NWRITE command"""
    cdef int [::1] nnum = np.empty(nnodes, ctypes.c_int)
    cdef double [:, ::1] nodes = np.empty((nnodes, 3), np.double)

    read_nblock_from_nwrite(filename, &nnum[0], &nodes[0, 0], nnodes)
    return np.array(nnum), np.array(nodes)


def write_array(filename, const double [::1] arr):
    cdef int nvalues = arr.size
    write_array_ascii(filename, &arr[0], nvalues)



import numpy as np
import numpy.typing as npt

def reset_midside(
    cellarr: npt.NDArray[np.int64],
    celltypes: npt.NDArray[np.uint8],
    offset: npt.NDArray[np.int64],
    pts: npt.NDArray[np.float64],
) -> None: ...



"""MAPDL archive reader."""

from mapdl_archive import examples  # noqa: F401
from mapdl_archive._version import __version__  # noqa: F401
from mapdl_archive.archive import (  # noqa: F401
    Archive,
    save_as_archive,
    write_cmblock,
    write_nblock,
)



"""Installation file for mapdl_archive."""

from io import open as io_open
import os

import numpy as np
from setuptools import Extension, setup

DEBUG = False
if DEBUG:
    extra_compile_args = ["-O0", "-g"]
else:
    extra_compile_args = ["/O2", "/w"] if os.name == "nt" else ["-O3", "-w"]


# Get version from version info
__version__ = None
this_file = os.path.dirname(__file__)
version_file = os.path.join(this_file, "mapdl_archive", "_version.py")
with io_open(version_file, mode="r") as fd:
    exec(fd.read())


setup(
    name="mapdl-archive",
    packages=["mapdl_archive", "mapdl_archive.examples"],
    version=__version__,
    description="Pythonic interface to MAPDL archive files.",
    long_description=open("README.rst").read(),
    long_description_content_type="text/x-rst",
    author="Alex Kaszynski",
    author_email="akascap@gmail.com",
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: MacOS",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    url="https://github.com/akaszynski/mapdl-archive",
    # Build cython modules
    include_dirs=[np.get_include()],
    ext_modules=[
        Extension(
            "mapdl_archive._relaxmidside",
            ["mapdl_archive/cython/_relaxmidside.pyx"],
            extra_compile_args=extra_compile_args,
            language="c",
        ),
        Extension(
            "mapdl_archive._archive",
            [
                "mapdl_archive/cython/_archive.pyx",
                "mapdl_archive/cython/archive.c",
            ],
            extra_compile_args=extra_compile_args,
            language="c",
        ),
        Extension(
            "mapdl_archive._reader",
            [
                "mapdl_archive/cython/_reader.pyx",
                "mapdl_archive/cython/reader.c",
                "mapdl_archive/cython/vtk_support.c",
            ],
            extra_compile_args=extra_compile_args,
            language="c",
        ),
    ],
    python_requires=">=3.8",
    keywords="vtk MAPDL ANSYS cdb",
    package_data={
        "mapdl_archive": ["py.typed"],
        "mapdl_archive.examples": [
            "TetBeam.cdb",
            "HexBeam.cdb",
            "sector.cdb",
        ],
    },
    install_requires=["pyvista>=0.41.1"],
)
